"""
Generate the vector features for pharmacophores from a rdkit conformer.

Adapted from rdkit:
    https://github.com/rdkit/rdkit/blob/master/rdkit/Chem/Features/FeatDirUtilsRD.py
    https://github.com/rdkit/rdkit/blob/master/rdkit/Chem/Features/ShowFeats.py

Changed to return anchor position and relative unit vector.
"""

from rdkit import Chem
from rdkit.Geometry.rdGeometry import Point3D

# pharmacophores
from rdkit.Chem.Features import FeatDirUtilsRD as FeatDirUtils

## Streamlined functions

def GetAromaticFeatVects(conf, featAtoms, featLoc, return_both: bool = False, scale=1.0):
    """
    Compute the direction vector for an aromatic feature

    Changed: only return one vector, process later for visualization and scoring

    Arguments
    ---------
        conf : a conformer
        featAtoms : list of atom IDs that make up the feature
        featLoc : location of the aromatic feature specified as point3d
        return_both : bool for whether to return both vectors or just one.
        scale : the size of the direction vector

    Returns
    -------
    Tuple
        list of anchor position(s) as rdkit Point3D
        list of relative unit vector(s) as rdkit Point3D
    """
    head = featLoc
    ats = [conf.GetAtomPosition(x) for x in featAtoms]

    v1 = ats[0] - head
    v2 = ats[1] - head
    norm1 = v1.CrossProduct(v2)
    norm1.Normalize()
    norm1 *= scale
    if return_both:
        norm2 = Point3D(0, 0, 0) - norm1
        return [head]*2, [norm1, norm2]
    else:
        return [head], [norm1]


def GetDonorFeatVects(conf, featAtoms, scale=1., exclude=[]):
    """
    Get vectors for hydrogen bond donors in the direction of the hydrogens.

    Arguments
    ---------
    conf : rdkit Mol object with a conformer.
    featAtoms : list containing rdkit Atom object of atom attributed as a donor.
    scale : float (default = 1.) length of direction vector.
    exclude : list of atom indices that should not be included as a donatable H.

    Returns
    -------
    Tuple
        list of anchor position(s) as rdkit Point3D or None
        list of relative unit vector(s) as rdkit Point3D or None
        list of neighboring hydrogens or None
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    mol = conf.GetOwningMol()
    cpt = conf.GetAtomPosition(aid)
    nbrs = mol.GetAtomWithIdx(aid).GetNeighbors()

    hyd_nbrs = [] # hydrogens
    vectors = []
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
            if nbr.GetIdx() in exclude:
                continue
            hyd_nbrs.append(nbr)
            hnid = nbr.GetIdx()
            vec = conf.GetAtomPosition(hnid)
            vec -= cpt
            vec.Normalize()
            vec *= scale
            vectors.append(vec)

    if hyd_nbrs != []:
        return [cpt]*len(vectors), vectors, hyd_nbrs
    else:
        return None, None, None


def GetAcceptorFeatVects(conf: Chem.rdchem.Mol,
                         featAtoms: Chem.rdchem.Atom,
                         scale: float = 1.0):
    """
    Get the anchor positions and relative unit vectors of an acceptor atom.

    Assumes HBA's are only O and N as defined by smarts_features.fdef.
    If HBA is not one of those, then it assumes the atom has one lone pair.

    Parameters
    ----------
    conf : Chem.Mol
        RDKit Mol object with a conformer.
    featAtoms : list
        List containing RDKit Atom object of atom attributed as an acceptor.
    scale : float, optional
        Length of direction vector. Default is 1.0.

    Returns
    -------
    tuple
        (list of anchor position(s) as RDKit Point3D or [None],
        list of relative unit vector(s) as RDKit Point3D or [None])
    """
    assert len(featAtoms) == 1
    atom_id = featAtoms[0]

    mol = conf.GetOwningMol()
    atom = mol.GetAtomWithIdx(atom_id)
    nbrs = atom.GetNeighbors()
    num_lone_pairs = 3
    if atom.GetAtomicNum() == 7: # N
        num_lone_pairs = 1
    elif atom.GetAtomicNum() == 8: # O
        num_lone_pairs = 2

    hydrogens = []
    heavy = []
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
            hydrogens.append(nbr)
        else:
            heavy.append(nbr)

    cpt = conf.GetAtomPosition(atom_id)

    if num_lone_pairs == 1:
        if len(nbrs) == 1: # linear
            # triple bond
            v1 = conf.GetAtomPosition(nbrs[0].GetIdx())
            v1 -= cpt
            v1.Normalize()
            v1 *= (-1.0 * scale)
            return [cpt], [v1]

        elif len(nbrs) == 2: # sp2
            v1 = FeatDirUtils._findAvgVec(conf, cpt, nbrs)
            v1 *= (-1.0 * scale)
            return [cpt], [v1]

        elif len(nbrs) == 3: # sp3
            cpt = conf.GetAtomPosition(atom_id)
            out = FeatDirUtils._GetTetrahedralFeatVect(conf, atom_id, scale)
            if out != ():
                v1 = out[0][1] - cpt # need to subtract out the center to be relative
                v1.Normalize()
                v1 *= scale
                return [cpt], [v1]
            else:
                # Means that this is planar so likely not an acceptor, remove it after
                return None, None

        else:
            return None, None

    elif num_lone_pairs == 2:
        heavy_nbr = heavy[0]
        if len(nbrs) == 1: # sp2
            for a in heavy_nbr.GetNeighbors():
                if a.GetIdx() != atom_id:
                    heavy_nbr_nbr = a # heavy atom's neighbor that isn't the acceptor
                    break

            pt1 = conf.GetAtomPosition(heavy_nbr_nbr.GetIdx())
            v1 = conf.GetAtomPosition(heavy_nbr.GetIdx())
            pt1 -= v1
            v1 -= cpt
            rotAxis = v1.CrossProduct(pt1)
            rotAxis.Normalize()
            bv1 = FeatDirUtils.ArbAxisRotation(120, rotAxis, v1)
            bv1.Normalize()
            bv1 *= scale
            bv2 = FeatDirUtils.ArbAxisRotation(-120, rotAxis, v1)
            bv2.Normalize()
            bv2 *= scale
            return [cpt]*2, [bv1, bv2]

        if len(nbrs) == 2: # sp3
            bvec = FeatDirUtils._findAvgVec(conf, cpt, nbrs)
            bvec *= (-1.0 * scale)
            # we will create two vectors by rotating bvec by half the tetrahedral angle in either directions
            v1 = conf.GetAtomPosition(nbrs[0].GetIdx())
            v1 -= cpt
            v2 = conf.GetAtomPosition(nbrs[1].GetIdx())
            v2 -= cpt
            rotAxis = v1 - v2
            rotAxis.Normalize()
            bv1 = FeatDirUtils.ArbAxisRotation(54.5, rotAxis, bvec)
            bv2 = FeatDirUtils.ArbAxisRotation(-54.5, rotAxis, bvec)
            bv1.Normalize()
            bv2.Normalize()
            bv1 *= scale
            bv2 *= scale
            return [cpt]*2, [bv1, bv2]

        else:
            return None, None

    elif num_lone_pairs == 3: # sp3 but do linear
        # Just do opposite of single bond (i.e., F)
        heavy_nbr = heavy[0]
        if len(hydrogens) == 0:
            v1 = conf.GetAtomPosition(heavy_nbr.GetIdx())
            v1 -= cpt
            v1.Normalize()
            v1 *= (-1.0 * scale)
            return [cpt], [v1]
    else:
        return None, None


def GetHalogenFeatVects(conf: Chem.rdchem.Mol,
                        featAtoms: Chem.rdchem.Atom,
                        scale: float = 1.0):
    """
    Get the anchor positions and relative unit vectors of a halogen atom.
    Assumes only one connection.

    Arguments
    ---------
    conf : rdkit Mol object with a conformer.
    featAtoms : list containing rdkit Atom object of atom attributed as an acceptor.
    scale : float (default = 1.) length of direction vector.

    Returns
    -------
    Tuple
        list of anchor position(s) as rdkit Point3D or [None]
        list of relative unit vector(s) as rdkit Point3D or [None]
    """
    assert len(featAtoms) == 1
    atom_id = featAtoms[0]

    mol = conf.GetOwningMol()
    atom = mol.GetAtomWithIdx(atom_id)
    nbrs = atom.GetNeighbors()

    cpt = conf.GetAtomPosition(atom_id)

    # Just do opposite of single bond
    heavy_nbr = nbrs[0]
    v1 = conf.GetAtomPosition(heavy_nbr.GetIdx())
    v1 -= cpt
    v1.Normalize()
    v1 *= (-1.0 * scale)
    return [cpt], [v1]


###################################################
## Modular functions from rdkit that were edited ##
## No longer used, but potentially more general ###
###################################################

# Hydrogen Bond Donors
def GetDonor1FeatVects_single(conf, featAtoms, scale=1.):
    """
    Get the direction vectors for Donor of type 1.
    Made to generate a single vector representation.

    This is a donor with one heavy atom. It is not clear where we should we should be putting the
    direction vector for this. It should probably be a cone. In this case we will just use the
    direction vector from the donor atom to the heavy atom.

    Changed: conditioning based on the number of hydrogens
    1. If 1 hydrogen, vector should point in the direction of the hydrogen.
    2. If 2 hydrogens, vector should point in a bisecting direction of the two hydrogens.
    3. If 3 hydrogens, point in the direction of the bond.

    Arguments
    ---------
    conf - rdkit Mol object with conformer
    featAtoms - list of atoms that are part of the feature
    scale - float for length of the direction vector (default = 1.0)

    Returns
    -------
    Tuple
        anchor position as rdkit Point3D or None
        relative unit vector(s) as rdkit Point3D or None
        list of hydrogen rdkit Atom objects
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    mol = conf.GetOwningMol()
    nbrs = mol.GetAtomWithIdx(aid).GetNeighbors()

    hyd_nbrs = [] # hydrogens
    heavy_nbr = -1 # hnbr in rdkit
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
            hyd_nbrs.append(nbr)
        else:
            heavy_nbr = nbr.GetIdx()

    cpt = conf.GetAtomPosition(aid)
    # Point in direction of hydrogen
    if len(hyd_nbrs) == 1:
        bvec = conf.GetAtomPosition(hyd_nbrs[0].GetIdx())
        bvec -= cpt
        bvec.Normalize()
        bvec *= scale
        return cpt, bvec, hyd_nbrs

    # Point in "average" (bisected) direction of 2 hydrogens
    elif len(hyd_nbrs) == 2:
        bvec = FeatDirUtils._findAvgVec(conf, cpt, hyd_nbrs)
        bvec *= scale
        # Removed conditional for generating 2 vectors for sp3 oxygen to maintain one average vector
        return cpt, bvec, hyd_nbrs

    # Otherwise, vector is the one from donor atom to heavy atom. 3 hydrogens
    cpt = conf.GetAtomPosition(aid)
    v1 = conf.GetAtomPosition(heavy_nbr)
    v1 -= cpt
    v1.Normalize()
    v1 *= (-1.0 * scale)
    return cpt, v1, hyd_nbrs


def GetDonor2FeatVects_single(conf, featAtoms, scale=1.):
    """
    Get the direction vectors for Donor of type 2.
    Made to generate a single vector representation.

    This is a donor with two heavy atoms as neighbors. The atom may are may not have
    hydrogen on it. Here are the situations with the neighbors that will be considered here
    1. two heavy atoms and two hydrogens: we will assume a sp3 arrangement here
    2. two heavy atoms and one hydrogen: this can either be sp2 or sp3
    3. two heavy atoms and no hydrogens

    Changed: conditioning based on the number of hydrogens
    1. For case 1, point in the direction bisecting the two hydrogens.
    2. For case 2, point in the direction of the hydrogen.
    3. For case 3, no changes.

    Arguments
    ---------
    conf : rdkit Mol object with conformer
    featAtoms : list of atoms that are part of the feature
    scale : float for length of the direction vector (default = 1.0)

    Returns
    -------
    Tuple
        anchor position as rdkit Point3D or None
        relative unit vector(s) as rdkit Point3D or None
        list of hydrogen rdkit Atom objects
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    mol = conf.GetOwningMol()
    cpt = conf.GetAtomPosition(aid)

    # find the two atoms that are neighbors of this atoms
    nbrs = list(mol.GetAtomWithIdx(aid).GetNeighbors())
    assert len(nbrs) >= 2

    hydrogens = []
    heavy = []
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
          hydrogens.append(nbr)
        else:
          heavy.append(nbr)

    # Case 3: not sure when this would be triggered
    if len(nbrs) == 2:
        # there should be no hydrogens in this case
        assert len(hydrogens) == 0
        # in this case the direction is the opposite of the average vector of the two neighbors
        bvec = FeatDirUtils._findAvgVec(conf, cpt, heavy)
        bvec *= (-1.0 * scale)
        return cpt, bvec, None

    # Case 2
    if len(nbrs) == 3:
        assert len(hydrogens) == 1
        # this is a little more tricky we have to check if the hydrogen is in the plane of the
        # two heavy atoms (i.e. sp2 arrangement) or out of plane (sp3 arrangement)
        # One of the directions will be from hydrogen atom to the heavy atom
        hid = hydrogens[0].GetIdx()
        bvec = conf.GetAtomPosition(hid)
        bvec -= cpt
        bvec.Normalize()
        bvec *= scale
        if FeatDirUtils._checkPlanarity(conf, cpt, nbrs, tol=1.0e-2):
            # only the hydrogen atom direction needs to be used
            return cpt, bvec, hydrogens

        # we have a non-planar configuration - we will assume sp3 and compute a second direction vector
        # Changed since we constrain to 1 vector
        return cpt, bvec, hydrogens

    # Case 1
    if len(nbrs) >= 4:
        # Changed -> take the bisecting vector (average) to only use 1 vector
        vecs = []
        for hid in hydrogens:
            hid = hid.GetIdx()
            bvec = conf.GetAtomPosition(hid)
            bvec -= cpt
            vecs.append(bvec.Normalize())
        bisec_vec = sum(*vecs)
        bisec_vec.Normalize()
        return cpt, bisec_vec, hydrogens
    return None, None, None


def GetDonor3FeatVects_single(conf, featAtoms, scale=1.0):
    """
    Get the direction vectors for Donor of type 3.
    Made to generate a single vector representation.

    This is a donor with three heavy atoms as neighbors. We will assume
    a tetrahedral arrangement of these neighbors. So the direction we are seeking
    is the last fourth arm of the sp3 arrangement

    Changed: Return anchor and relative unit vector tuple

    Arguments
    ---------
    conf : rdkit Mol object with conformer
    featAtoms : list of atoms that are part of the feature
    scale : float for length of the direction vector (default = 1.0)

    Returns
    -------
    Tuple
        anchor position as rdkit Point3D or None
        relative unit vector(s) as rdkit Point3D or None
        list of hydrogen rdkit Atom objects
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    cpt = conf.GetAtomPosition(aid)
    out = FeatDirUtils._GetTetrahedralFeatVect(conf, aid, scale)
    nbrs = conf.GetAtomWithIdx(aid).GetNeighbors()
    hydrogens = []
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
            hydrogens.append(nbr)
    if out != ():
        if hydrogens == []:
            hydrogens = None
        return cpt, out[0][1] - cpt, hydrogens # need to subtract out the center
    else:
        # Means that this is planar so likely not an donor, remove it after
        return None, None, None


# Hydrogen bond acceptors
def GetAcceptor1FeatVects_single(conf, featAtoms, scale=1.):
    """
    Get the direction vectors for Acceptor of type 1 (single vector representation).

    This is an acceptor with one heavy atom neighbor. There are two possibilities:

    - The bond to the heavy atom is a single bond (e.g. CO): We use the inversion
      of this bond direction and mark it as a 'cone'.
    - The bond to the heavy atom is a double bond (e.g. C=O): We have two possible
      directions except in some special cases (e.g. SO2) where we use bond direction.

    Notes
    -----
    Modified to condition on the number of hydrogens with methanamine fix:

    - Case 1: If one hydrogen, vector points in the opposite direction of the bisection
      of the acute angle formed by the heavy-acceptor-hydrogen. If two hydrogens,
      assume sp3 and project in that lone-pair direction. If not tetrahedral,
      return None.
    - Case 2: Return the bisecting vector of the two lone-pairs.

    Parameters
    ----------
    conf : Chem.Mol
        RDKit Mol object with conformer.
    featAtoms : list
        List of atoms that are part of the feature.
    scale : float, optional
        Length of the direction vector. Default is 1.0.

    Returns
    -------
    tuple
        (anchor position as RDKit Point3D or None,
        relative unit vector(s) as RDKit Point3D or None)
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    mol = conf.GetOwningMol()
    nbrs = mol.GetAtomWithIdx(aid).GetNeighbors()

    cpt = conf.GetAtomPosition(aid)

    hyd_nbrs = [] # hydrogens
    heavyAt = -1
    # find the adjacent heavy atom and hydrogens
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
            hyd_nbrs.append(nbr)
        else:
            heavyAt = nbr

    # before this was ``>`` which doesn't make sense
    singleBnd = mol.GetBondBetweenAtoms(aid, heavyAt.GetIdx()).GetBondType() == Chem.BondType.SINGLE

    # special scale - if the heavy atom is a sulfur (we should proabably check phosphorous as well)
    sulfur = heavyAt.GetAtomicNum() == 16

    methanamine = mol.GetBondBetweenAtoms(aid, heavyAt.GetIdx()).GetBondType() == Chem.BondType.DOUBLE and len(hyd_nbrs) == 1

    if singleBnd or sulfur or methanamine:
        if len(hyd_nbrs) == 1:
            bvec = FeatDirUtils._findAvgVec(conf, cpt, [heavyAt, hyd_nbrs[0]])
            bvec *= (-1 * scale)
            return cpt, bvec

        elif len(hyd_nbrs) == 2:
            out = FeatDirUtils._GetTetrahedralFeatVect(conf, aid, scale)
            if out != ():
                return cpt, out[0][1] - cpt # need to subtract out the center
            else:
                # Means that this is planar so likely not an acceptor, remove it after
                return None, None

    # Changed -> Assume sp2 (like the original code) but rather than get the two
    #  lone pair directions just do the vector along the double bond axis to use only one vector
    v1 = conf.GetAtomPosition(heavyAt.GetIdx())
    v1 -= cpt
    v1.Normalize()
    v1 *= (-1.0 * scale)
    return cpt, v1


def GetAcceptor2FeatVects_single(conf, featAtoms, scale=1.):
    """
    Get the direction vectors for Acceptor of type 2.
    Made to generate a single vector representation.

    This is the acceptor with two adjacent heavy atoms. We will special case a few things here.
    If the acceptor atom is an oxygen we will assume a sp3 hybridization
    the acceptor directions (two of them)
    reflect that configurations. Otherwise the direction vector in plane with the neighboring
    heavy atoms

    Changed: Only generate one vector
    Rather than generating 2 vectors for sp3 oxygen, just keep the average vector.

    Arguments
    ---------
    conf : rdkit Mol object with conformer
    featAtoms : list of atoms that are part of the feature
    scale : float for length of the direction vector (default = 1.0)

    Returns
    -------
    Tuple
        anchor position as rdkit Point3D or None
        relative unit vector(s) as rdkit Point3D or None
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    cpt = conf.GetAtomPosition(aid)

    mol = conf.GetOwningMol()
    nbrs = list(mol.GetAtomWithIdx(aid).GetNeighbors())
    hydrogens = []
    heavy = []
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
          hydrogens.append(nbr)
        else:
          heavy.append(nbr)

    if len(hydrogens) == 0:
        bvec = FeatDirUtils._findAvgVec(conf, cpt, heavy)
    elif len(hydrogens) == 1:
        bvec = FeatDirUtils._findAvgVec(conf, cpt, nbrs)
    bvec *= (-1.0 * scale)

    # Changed -- Removed conditional for generating 2 vectors for sp3 oxygen to maintain one average vector
    return cpt, bvec


def GetAcceptor3FeatVects_single(conf, featAtoms, scale=1.0):
    """
    Get the direction vectors for Donor of type 3.
    Made to generate a single vector representation.

    This is a donor with three heavy atoms as neighbors. We will assume
    a tetrahedral arrangement of these neighbors. So the direction we are seeking
    is the last fourth arm of the sp3 arrangement

    Changed: to return anchor and relative unit vector tuple

    Arguments
    ---------
    conf : rdkit Mol object with conformer
    featAtoms : list of atoms that are part of the feature
    scale : float for length of the direction vector (default = 1.0)

    Returns
    -------
    Tuple
        anchor position as rdkit Point3D or None
        relative unit vector(s) as rdkit Point3D or None
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    cpt = conf.GetAtomPosition(aid)
    out = FeatDirUtils._GetTetrahedralFeatVect(conf, aid, scale)
    if out != ():
        return cpt, out[0][1] - cpt # need to subtract out the center to be relative
    else:
        # Means that this is planar so likely not an acceptor, remove it after
        return None, None


################### MULTI-VECTOR ##################
## THIS IS MORE TRUTHFUL TO RDKIT IMPLEMENTATION ##

# Hydrogen bond acceptors
def GetAcceptor1FeatVects(conf, featAtoms, scale=1.):
    """
    Get the direction vectors for Acceptor of type 1 (multi-vector representation).

    This is an acceptor with one heavy atom neighbor. There are two possibilities:

    - The bond to the heavy atom is a single bond (e.g. CO): We use the inversion
      of this bond direction and mark it as a 'cone'.
    - The bond to the heavy atom is a double bond (e.g. C=O): We have two possible
      directions except in some special cases (e.g. SO2) where we use bond direction.

    Notes
    -----
    Modified to change return format, with fixes for methanamine and two vectors
    for hydroxyls.

    Parameters
    ----------
    conf : Chem.Mol
        RDKit Mol object with a conformer.
    featAtoms : list
        List containing RDKit Atom object of atom attributed as an acceptor.
    scale : float, optional
        Length of direction vector. Default is 1.0.

    Returns
    -------
    tuple
        (list of anchor position(s) as RDKit Point3D or None,
        list of relative unit vector(s) as RDKit Point3D or None)
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    mol = conf.GetOwningMol()
    nbrs = mol.GetAtomWithIdx(aid).GetNeighbors()

    cpt = conf.GetAtomPosition(aid)

    hyd_nbrs = [] # hydrogens
    heavyAt = -1
    # find the adjacent heavy atom and hydrogens
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
            hyd_nbrs.append(nbr)
        else:
            heavyAt = nbr

    # before this was ``>`` which doesn't make sense
    singleBnd = mol.GetBondBetweenAtoms(aid, heavyAt.GetIdx()).GetBondType() == Chem.BondType.SINGLE

    # special scale - if the heavy atom is a sulfur (we should proabably check phosphorous as well)
    sulfur = heavyAt.GetAtomicNum() == 16

    methanamine = mol.GetBondBetweenAtoms(aid, heavyAt.GetIdx()).GetBondType() == Chem.BondType.DOUBLE and len(hyd_nbrs) == 1

    if singleBnd or sulfur or methanamine:
        if len(hyd_nbrs) == 0 or sulfur:
            v1 = conf.GetAtomPosition(heavyAt.GetIdx())
            v1 -= cpt
            v1.Normalize()
            v1 *= (-1.0 * scale)
            return [cpt], [v1]

        elif len(hyd_nbrs) == 1 and methanamine:
            bvec = FeatDirUtils._findAvgVec(conf, cpt, [heavyAt, hyd_nbrs[0]])
            bvec *= (-1 * scale)
            return [cpt], [bvec]

        elif len(hyd_nbrs) == 1 and singleBnd:
            # hydroxyl group
            bvec = FeatDirUtils._findAvgVec(conf, cpt, nbrs)
            bvec *= (-1.0 * scale)

            # assume sp3
            # we will create two vectors by rotating bvec by half the tetrahedral angle in either directions
            v1 = conf.GetAtomPosition(hyd_nbrs[0].GetIdx())
            v1 -= cpt
            v2 = conf.GetAtomPosition(heavyAt.GetIdx())
            v2 -= cpt
            rotAxis = v1 - v2
            rotAxis.Normalize()
            bv1 = FeatDirUtils.ArbAxisRotation(54.5, rotAxis, bvec)
            bv2 = FeatDirUtils.ArbAxisRotation(-54.5, rotAxis, bvec)
            return [cpt]*2, [bv1, bv2]

        elif len(hyd_nbrs) == 2:
            out = FeatDirUtils._GetTetrahedralFeatVect(conf, aid, scale)
            if out != ():
                return [cpt], [out[0][1] - cpt] # need to subtract out the center
            else:
                # Means that this is planar so likely not an acceptor, remove it after
                return None, None

    # ok in this case we will assume that
    # heavy atom is sp2 hybridized and the direction vectors (two of them)
    # are in the same plane, we will find this plane by looking for one
    # of the neighbors of the heavy atom
    hvNbrs = heavyAt.GetNeighbors()
    hvNbr = -1
    for nbr in hvNbrs:
        if nbr.GetIdx() != aid:
            hvNbr = nbr
            break

    pt1 = conf.GetAtomPosition(hvNbr.GetIdx())
    v1 = conf.GetAtomPosition(heavyAt.GetIdx())
    pt1 -= v1
    v1 -= cpt
    rotAxis = v1.CrossProduct(pt1)
    rotAxis.Normalize()
    bv1 = FeatDirUtils.ArbAxisRotation(120, rotAxis, v1)
    bv1.Normalize()
    bv1 *= scale
    bv2 = FeatDirUtils.ArbAxisRotation(-120, rotAxis, v1)
    bv2.Normalize()
    bv2 *= scale
    return [cpt]*2, [bv1, bv2]


def GetAcceptor2FeatVects(conf, featAtoms, scale=1.):
    """
    Get the direction vectors for Acceptor of type 2.
    Made to generate a single vector representation.

    This is the acceptor with two adjacent heavy atoms. We will special case a few things here.
    If the acceptor atom is an oxygen we will assume a sp3 hybridization
    the acceptor directions (two of them)
    reflect that configurations. Otherwise the direction vector in plane with the neighboring
    heavy atoms

    Changed: return format

    Arguments
    ---------
    conf : rdkit Mol object with a conformer.
    featAtoms : list containing rdkit Atom object of atom attributed as an acceptor.
    scale : float (default = 1.) length of direction vector.

    Returns
    -------
    Tuple
        list of anchor position(s) as rdkit Point3D or None
        list of relative unit vector(s) as rdkit Point3D or None
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    cpt = conf.GetAtomPosition(aid)

    mol = conf.GetOwningMol()
    nbrs = list(mol.GetAtomWithIdx(aid).GetNeighbors())
    hydrogens = []
    heavy = []
    for nbr in nbrs:
        if nbr.GetAtomicNum() == 1:
          hydrogens.append(nbr)
        else:
          heavy.append(nbr)

    if len(hydrogens) == 0:
        bvec = FeatDirUtils._findAvgVec(conf, cpt, heavy)
    elif len(hydrogens) == 1:
        bvec = FeatDirUtils._findAvgVec(conf, cpt, nbrs)
    bvec *= (-1.0 * scale)

    if mol.GetAtomWithIdx(aid).GetAtomicNum() == 8:
        # assume sp3
        # we will create two vectors by rotating bvec by half the tetrahedral angle in either directions
        v1 = conf.GetAtomPosition(heavy[0].GetIdx())
        v1 -= cpt
        v2 = conf.GetAtomPosition(heavy[1].GetIdx())
        v2 -= cpt
        rotAxis = v1 - v2
        rotAxis.Normalize()
        bv1 = FeatDirUtils.ArbAxisRotation(54.5, rotAxis, bvec)
        bv2 = FeatDirUtils.ArbAxisRotation(-54.5, rotAxis, bvec)
        return [cpt]*2, [bv1, bv2]
    return [cpt], [bvec]


def GetAcceptor3FeatVects(conf, featAtoms, scale=1.0):
    """
    Get the direction vectors for Donor of type 3.
    Made to generate a single vector representation.

    This is a donor with three heavy atoms as neighbors. We will assume
    a tetrahedral arrangement of these neighbors. So the direction we are seeking
    is the last fourth arm of the sp3 arrangement

    Changed: return format

    Arguments
    ---------
    conf : rdkit Mol object with a conformer.
    featAtoms : list containing rdkit Atom object of atom attributed as an acceptor.
    scale : float (default = 1.) length of direction vector.

    Returns
    -------
    Tuple
        list of anchor position(s) as rdkit Point3D or None
        list of relative unit vector(s) as rdkit Point3D or None
    """
    assert len(featAtoms) == 1
    aid = featAtoms[0]
    cpt = conf.GetAtomPosition(aid)
    out = FeatDirUtils._GetTetrahedralFeatVect(conf, aid, scale)
    if out != ():
        return [cpt], [out[0][1] - cpt] # need to subtract out the center to be relative
    else:
        # Means that this is planar so likely not an acceptor, remove it after
        return None, None



"""
Module contains docking evaluation pipeline and target information.
"""

from shepherd_score.evaluations.docking.targets import docking_target_info
from shepherd_score.evaluations.docking.pipelines import DockingEvalPipeline
from shepherd_score.evaluations.docking.pipelines import run_docking_evaluation
from shepherd_score.evaluations.docking.pipelines import run_docking_evaluation_from_smiles
from shepherd_score.evaluations.docking.pipelines import run_docking_benchmark

__all__ = [
    'docking_target_info',
    'DockingEvalPipeline',
    'run_docking_evaluation',
    'run_docking_evaluation_from_smiles',
    'run_docking_benchmark',
]



"""
Unit tests for alignment and related fucntions.
"""
import pytest
import numpy as np
import torch
from .utils import _configure_jax_platform
from shepherd_score.alignment_utils.se3_np import (
    quaternions_to_rotation_matrix_np,
    get_SE3_transform_np,
    apply_SE3_transform_np,
)
from shepherd_score.alignment_utils.se3 import (
    quaternions_to_rotation_matrix,
    get_SE3_transform,
    apply_SE3_transform,
)


# Attempt to import JAX and related modules
JAX_AVAILABLE = False
jnp = None
quaternions_to_rotation_matrix_jax = None
get_SE3_transform_jax = None
apply_SE3_transform_jax = None

try:
    # Configure JAX platform before import to avoid GPU initialization errors
    _gpu_detected = _configure_jax_platform()

    import jax.numpy as jnp
    from shepherd_score.alignment_utils.se3_jax import (
        quaternions_to_rotation_matrix_jax,
        get_SE3_transform_jax,
        apply_SE3_transform_jax
    )

    JAX_AVAILABLE = True
except ImportError:
    # JAX not available - JAX-specific tests will be skipped
    pass


class TestSE3:
    """
    Test SE(3) related transformations.
    Checks all versions: NumPy, Jax, Torch (batched and unbatched)
    """
    ex_quaternion = np.array([5., 0.2, -3.4, 0.4])
    # Numbers from from https://www.andre-gaschler.com/rotationconverter/
    sol_rot_matrix = np.array([
        [0.3623504, -0.1458107, -0.9205658],
        [0.0718172,  0.9891186, -0.1284004],
        [0.9292709, -0.0195865,  0.3688792]
    ])

    ex_se3_params = np.array([5., 0.2, -3.4, 0.4, 4., 5., 6.])
    sol_se3_transform = np.array([
        [0.3623504, -0.1458107, -0.9205658, 4.],
        [0.0718172,  0.9891186, -0.1284004, 5.],
        [0.9292709, -0.0195865,  0.3688792, 6.],
        [0.       , 0.        , 0.        , 1.]
    ])

    ex_set_of_points = np.array([
        [1., 2., 3.],
        [-1., 2., -3.],
        [0.4, -0.4, -0.5]
    ])
    sol_points_transformed = (sol_se3_transform[:3, :3] @ ex_set_of_points.T).T + sol_se3_transform[:3, 3]

    # NUMPY functions
    def test_quaternion_to_rotation_matrix_np(self):
        """ Test quaternions_to_rotation_matrix_np """
        out_rot_matrix = quaternions_to_rotation_matrix_np(self.ex_quaternion)
        assert np.allclose(out_rot_matrix, self.sol_rot_matrix)

    def test_se3_matrix_from_params_np(self):
        """ Test get_SE3_transform_np """
        out_se3_matrix = get_SE3_transform_np(self.ex_se3_params)
        assert np.allclose(out_se3_matrix, self.sol_se3_transform)

    def test_apply_se3_transform_np(self):
        """ Test apply_SE3_transform_np  """
        out_transformed = apply_SE3_transform_np(self.ex_set_of_points, self.sol_se3_transform)
        assert np.allclose(out_transformed, self.sol_points_transformed)

    # TORCH functions (single instance)
    def test_quaternion_to_rotation_matrix_torch(self):
        """ Test quaternions_to_rotation_matrix (Torch) """
        out_rot_matrix = quaternions_to_rotation_matrix(torch.Tensor(self.ex_quaternion))
        assert torch.allclose(out_rot_matrix, torch.Tensor(self.sol_rot_matrix))

    def test_se3_matrix_from_params_torch(self):
        """ Test get_SE3_transform (torch) """
        out_se3_matrix = get_SE3_transform(torch.Tensor(self.ex_se3_params))
        assert torch.allclose(out_se3_matrix, torch.Tensor(self.sol_se3_transform))

    def test_apply_se3_transform_torch(self):
        """ Test apply_SE3_transform torch  """
        out_transformed = apply_SE3_transform(torch.Tensor(self.ex_set_of_points), torch.Tensor(self.sol_se3_transform))
        assert torch.allclose(out_transformed, torch.Tensor(self.sol_points_transformed))

    # TORCH functions (batched)
    def test_quaternion_to_rotation_matrix_torch_batched(self):
        """ Test quaternions_to_rotation_matrix (Torch batched) """
        out_rot_matrix = quaternions_to_rotation_matrix(torch.Tensor(self.ex_quaternion).repeat((2, 1)))
        assert torch.allclose(out_rot_matrix, torch.Tensor(self.sol_rot_matrix).repeat((2,1,1)))

    def test_se3_matrix_from_params_torch_batched(self):
        """ Test get_SE3_transform (torch batched) """
        out_se3_matrix = get_SE3_transform(torch.Tensor(self.ex_se3_params).repeat((2,1)))
        assert torch.allclose(out_se3_matrix, torch.Tensor(self.sol_se3_transform).repeat((2, 1, 1)))

    def test_apply_se3_transform_torch_batched(self):
        """ Test apply_SE3_transform (torch batched)  """
        points_repeated = torch.Tensor(self.ex_set_of_points).repeat((2, 1, 1))
        transform_repeated = torch.Tensor(self.sol_se3_transform).repeat((2, 1, 1))
        out_transformed = apply_SE3_transform(points_repeated, transform_repeated)
        sol_repeated = torch.Tensor(self.sol_points_transformed).repeat((2, 1, 1))
        assert torch.allclose(out_transformed, sol_repeated)

    # Jax functions (single instance)
    @pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX is not installed")
    def test_quaternion_to_rotation_matrix_jax(self):
        """ Test quaternions_to_rotation_matrix (Jax) """
        out_rot_matrix = quaternions_to_rotation_matrix_jax(jnp.array(self.ex_quaternion))
        assert jnp.allclose(out_rot_matrix, jnp.array(self.sol_rot_matrix))

    @pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX is not installed")
    def test_se3_matrix_from_params_jax(self):
        """ Test get_SE3_transform (Jax) """
        out_se3_matrix = get_SE3_transform_jax(jnp.array(self.ex_se3_params))
        assert jnp.allclose(out_se3_matrix, jnp.array(self.sol_se3_transform))

    @pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX is not installed")
    def test_apply_se3_transform_jax(self):
        """ Test apply_SE3_transform (Jax)  """
        out_transformed = apply_SE3_transform_jax(jnp.array(self.ex_set_of_points), jnp.array(self.sol_se3_transform))
        assert jnp.allclose(out_transformed, jnp.array(self.sol_points_transformed))



class TestPCA:
    """
    Test principal component analysis
    """



"""
Module contains `Objective` class used for computing similarity scores between reference and fit
molecules.
"""

from typing import Union, Optional, List, Tuple
import os
from tqdm import tqdm

import numpy as np
from rdkit import Chem
import open3d # noqa: F401 (imported because sometimes order matters)
import torch

from shepherd_score.container import Molecule, MoleculePair
from shepherd_score.score.gaussian_overlap_np import get_overlap_np
from shepherd_score.score.gaussian_overlap import get_overlap
from shepherd_score.score.electrostatic_scoring_np import get_overlap_esp_np
from shepherd_score.score.electrostatic_scoring import get_overlap_esp
from shepherd_score.score.pharmacophore_scoring_np import get_overlap_pharm_np
from shepherd_score.alignment_utils.se3_np import apply_SE3_transform_np, apply_SO3_transform_np
from shepherd_score.conformer_generation import embed_conformer_from_smiles, optimize_conformer_with_xtb
from shepherd_score.conformer_generation import generate_opt_conformers, generate_opt_conformers_xtb, charges_from_single_point_conformer_with_xtb
from shepherd_score.score.constants import ALPHA

TMPDIR = '.'
if 'TMPDIR' in os.environ:
    TMPDIR = os.environ['TMPDIR']


class GeneralObjective:
    """
    Objective class that computes similarity scores of fit molecules given a reference molecule.
    """
    def __init__(self,
                 ref_mol: Union[str, Chem.Mol],
                 rep: str = 'esp',
                 xtb_opt: bool = False,
                 num_points: Optional[int] = None,
                 use_vol: bool = False,
                 pharm_multi_vector: Optional[bool] = None,
                 solvent: Optional[str] = None,
                 num_processes: int = 1):
        """
        Generalized objective class that computes similarity scores of fit molecules given a
        reference.
        Constructor -- sets up the reference molecule and relevant representations.

        Arguments
        ---------
        ref_mol : str or Chem.Mol representing reference molecule.
        rep : str for representation chosen from ('shape', 'esp', 'pharm')
        xtb_opt : bool to optimize conformers with xtb or just use EKTG embedding and MMFF opt.
        num_points : Optional[int] if `rep` is 'shape' or 'esp' this is necessary to generate the
            surface point cloud.
        use_vol : bool (default = False) toggle to True if you want to use volumetric scoring.
            Ignores `num_points` if True.
        pharm_multi_vector : Optional[bool] whether to represent pharmacophores with mulitple
            vectors (if applicable) or a single, averaged vector.
        solvent : Optional[str] solvent to use if optimizing with xtb. Default (None) is gas phase.
        num_processes : int (default = 1) Number of proccesses to use during xtb optimization
        """
        self.ref_partial_charges = None
        self.representation = rep.lower()
        self.xtb_opt = xtb_opt
        self.solvent = solvent
        self.use_vol = use_vol
        self.ref_molec = None

        self.buffer = {} # stores past SMILES and their similarity scores

        self.num_points = num_points
        if self.use_vol:
            self.alpha = 0.81
            self.num_points = None
        else:
            self.alpha = ALPHA(self.num_points) if self.num_points is not None else None

        self.lam = 0.1 if self.use_vol else 0.3

        self.pharm_multi_vector = pharm_multi_vector

        self.num_processes = num_processes

        self.sim_score_distr_with_resample = np.array([1.])
        self.sim_score_upper_bound = 1.

        if self.representation == 'shape':
            pass
        elif self.representation in ('electrostatics', 'esp'):
            self.representation = 'esp'
        elif self.representation in ('pharmacophore', 'pharmacophores', 'pharm'):
            self.representation = 'pharm'
        else:
            raise ValueError(f'Please enter a valid key for `rep`. "{rep}" was given')

        if self.representation in ('shape', 'esp') and self.num_points is None and not self.use_vol:
            raise ValueError('Either `use_vol` must be True or `num_points` must be supplied for surface point cloud.')
        if self.representation == 'pharm' and self.pharm_multi_vector is None:
            raise ValueError(f'`pharm_multi_vector` must be supplied for surface point cloud. {pharm_multi_vector} was given.')

        if isinstance(ref_mol, str):
            ref_mol = embed_conformer_from_smiles(ref_mol, attempts=50, MMFF_optimize=True)
            if self.xtb_opt:
                ref_mol, _, self.ref_partial_charges = optimize_conformer_with_xtb(
                    ref_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    temp_dir=TMPDIR
                )
            else:
                self.ref_partial_charges = np.array(charges_from_single_point_conformer_with_xtb(
                    ref_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    temp_dir=TMPDIR
                ))
        elif isinstance(ref_mol, Chem.Mol):
            try:
                ref_mol.GetConformer()
                has_conformer = True
            except Exception:
                has_conformer = False

            if not has_conformer:
                ref_mol = embed_conformer_from_smiles(ref_mol, attempts=50, MMFF_optimize=True)
                if self.xtb_opt:
                    ref_mol, _, self.ref_partial_charges = optimize_conformer_with_xtb(
                        ref_mol,
                        solvent=self.solvent,
                        num_cores=self.num_processes,
                        temp_dir=TMPDIR
                    )
            if self.ref_partial_charges is None:
                self.ref_partial_charges = np.array(charges_from_single_point_conformer_with_xtb(
                    ref_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    temp_dir=TMPDIR
                ))
        else:
            raise ValueError(f'`ref_molec` must be str or Chem.Mol object. Instead {type(ref_mol)} was given.')

        # Get Molecule objects
        if self.representation in ('shape', 'esp'):
            self.ref_molec = Molecule(mol=ref_mol,
                                      num_surf_points=self.num_points,
                                      partial_charges=self.ref_partial_charges)

        elif self.representation == 'pharm':
            self.ref_molec = Molecule(mol=ref_mol,
                                      pharm_multi_vector=self.pharm_multi_vector)

        if self.representation in ('shape', 'esp') and not self.use_vol:
            self.sim_score_distr_with_resample = self.resampling_surf_scores()
            self.sim_score_upper_bound = max(self.sim_score_distr_with_resample)

        self.buffer[Chem.MolToSmiles(self.ref_molec.mol)] = self.sim_score_upper_bound


    def resampling_surf_scores(self) -> np.ndarray:
        """
        Capture distribution of surface similarity or surface ESP scores caused by resampling
        surface.

        Returns
        -------
        score_distr : np.ndarray (if not relevant)
        """
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        if self.num_points >=150 and torch.cuda.is_available():
            use_torch = True
        else:
            use_torch = False

        num_repeats = 5

        score_distr = np.empty(num_repeats)
        if self.representation == 'shape':
            for i in range(num_repeats):
                molec = Molecule(mol=self.ref_molec.mol,
                                 num_surf_points=self.ref_molec.num_surf_points,
                                 probe_radius=self.ref_molec.probe_radius,
                                 partial_charges=self.ref_molec.partial_charges)
                if use_torch:
                    score_distr[i] = get_overlap(
                        centers_1=torch.from_numpy(self.ref_molec.surf_pos).to(torch.float32).to(device),
                        centers_2=torch.from_numpy(molec.surf_pos).to(torch.float32).to(device),
                        alpha=self.alpha
                    ).cpu().numpy()
                else:
                    score_distr[i] = get_overlap_np(
                        centers_1=self.ref_molec.surf_pos,
                        centers_2=molec.surf_pos,
                        alpha=self.alpha
                    )

        elif self.representation == 'esp':
            for i in range(num_repeats):
                molec = Molecule(mol=self.ref_molec.mol,
                                 num_surf_points=self.ref_molec.num_surf_points,
                                 probe_radius=self.ref_molec.probe_radius,
                                 partial_charges=self.ref_molec.partial_charges)
                if use_torch:
                    score_distr[i] = get_overlap_esp(
                        centers_1=torch.from_numpy(self.ref_molec.surf_pos).to(torch.float32).to(device),
                        centers_2=torch.from_numpy(molec.surf_pos).to(torch.float32).to(device),
                        charges_1=torch.from_numpy(self.ref_molec.surf_esp).to(torch.float32).to(device),
                        charges_2=torch.from_numpy(molec.surf_esp).to(torch.float32).to(device),
                        alpha=self.alpha,
                        lam=self.lam
                    ).cpu().numpy()
                else:
                    score_distr[i] = get_overlap_esp_np(
                        centers_1=self.ref_molec.surf_pos,
                        centers_2=molec.surf_pos,
                        charges_1=self.ref_molec.surf_esp,
                        charges_2=molec.surf_esp,
                        alpha=self.alpha,
                        lam=self.lam
                    )
        else:
            score_distr = np.array([1.]*num_repeats)

        return score_distr


    def _score(self,
               fit_mol: Chem.Mol,
               fit_partial_charges: Optional[np.ndarray] = None,
               trans_init: bool = False,
               use_jax: bool = False) -> float:
        """
        Align and score a fit molecule to the reference molecule.

        Arguments
        ---------
        fit_mol : Chem.Mol fit molecule to compare to reference.
        fit_partial_charges : np.ndarray of partial charges for each atom
        trans_init : bool (default = False) Whether to initially translate fit COM to each ref atom
            during alignment.
        use_jax : bool (default = False) uses torch if False

        Returns
        -------
        float : similarity score
        """
        if fit_partial_charges is not None:
            fit_partial_charges = np.array(fit_partial_charges)
        if self.representation in ('shape', 'electrostatics', 'esp'):
            fit_molec = Molecule(mol=fit_mol,
                                 num_surf_points=self.num_points,
                                 partial_charges=fit_partial_charges)

        elif self.representation in ('pharmacophore', 'pharmacophores', 'pharm', 'p'):
            fit_molec = Molecule(mol=fit_mol,
                                 pharm_multi_vector=self.pharm_multi_vector)

        molec_pair = MoleculePair(self.ref_molec, fit_molec, num_surf_points=self.num_points)

        if self.representation == 'shape':
            if self.use_vol:
                molec_pair.align_with_vol(no_H=True, trans_init=trans_init, use_jax=use_jax)
                score = molec_pair.sim_aligned_vol_noH
            else:
                molec_pair.align_with_surf(alpha=self.alpha, trans_init=trans_init, use_jax=use_jax)
                score = molec_pair.sim_aligned_surf

        elif self.representation == 'esp':
            if self.use_vol:
                molec_pair.align_with_vol_esp(lam=self.lam, no_H=True, trans_init=trans_init, use_jax=use_jax)
                score = molec_pair.sim_aligned_vol_esp_noH
            else:
                molec_pair.align_with_esp(self.alpha,
                                        lam=self.lam,
                                        trans_init=trans_init,
                                        use_jax=use_jax)
                score = molec_pair.sim_aligned_esp

        elif self.representation == 'pharm':
            molec_pair.align_with_pharm(similarity='tanimoto',
                                        extended_points=False,
                                        only_extended=False,
                                        trans_init=trans_init,
                                        use_jax=False)
            score = molec_pair.sim_aligned_pharm

        return score


    def score(self,
              smiles: str,
              num_conformers: int = 1,
              trans_init: bool = False,
              use_jax: bool = False) -> float:
        """
        Align and score a fit molecule to the reference molecule.

        Arguments
        ---------
        smiles : str SMILES string of molecule to compare to reference.
        num_conformers : int (default = 1) Max number of conformers to generate and score.
        trans_init : bool (default = False) Whether to initially translate fit COM to each ref atom
            during alignment.
        use_jax : bool (default = False) uses torch if False

        Returns
        -------
        float : similarity score (or max score if num_conformers > 1)
        """
        fit_partial_charges = None
        if num_conformers == 1:
            fit_mol = embed_conformer_from_smiles(smiles, attempts=50, MMFF_optimize=True)

            if self.xtb_opt:
                fit_mol, _, fit_partial_charges = optimize_conformer_with_xtb(
                    fit_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    temp_dir=TMPDIR
                )
            if fit_partial_charges is None:
                fit_partial_charges = charges_from_single_point_conformer_with_xtb(
                    fit_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    temp_dir=TMPDIR
                )
            score = self._score(fit_mol,
                                trans_init=trans_init,
                                use_jax=use_jax)
            return score

        if self.xtb_opt:
            fit_mols, fit_partial_charges = generate_opt_conformers_xtb(smiles, MMFF_optimize=True, verbose=False)
        else:
            fit_mols = generate_opt_conformers(smiles, MMFF_optimize=True, verbose=False)
            fit_partial_charges = []
            for m in fit_mols:
                fit_partial_charges.append(charges_from_single_point_conformer_with_xtb(
                    conformer=m,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    temp_dir=TMPDIR
                ))

        scores = []
        for i, fit_mol in enumerate(fit_mols):
            scores.append(self._score(fit_mol,
                                      fit_partial_charges=fit_partial_charges[i],
                                      trans_init=trans_init,
                                      use_jax=use_jax))

        return max(scores)


    def score_multiple(self,
                       smiles: List[str],
                       num_conformers: int = 1,
                       trans_init: bool = False,
                       use_jax: bool = False) -> List[float]:
        """
        Aligns and scores multiple fit SMILES.

        Arguments
        ---------
        smiles : List[str] of SMILES strings of molecules to compare to reference.
        num_conformers : int (default = 1) Max number of conformers to generate and score.
        trans_init : bool (default = False) Whether to initially translate fit COM to each ref atom
            during alignment.
        use_jax : bool (default = False) uses torch if False

        Returns
        -------
        List[float] : similarity scores. Returns a value of -1 if Objective.score fails for any reason.
        """
        scores = []
        for smi in smiles:
            smi = Chem.CanonSmiles(smi)
            if smi in self.buffer:
                # skip if we've already computed it
                scores.append(self.buffer[smi])
            else:
                try:
                    scores.append(self.score(smi,
                                             num_conformers=num_conformers,
                                             trans_init=trans_init,
                                             use_jax=use_jax))
                except Exception:
                    scores.append(-1.)

            self.buffer[smi] = scores[-1] # store {smiles : score}

        return scores



class Objective:
    """
    Objective class that computes similarity scores of fit molecules given a reference molecule.
    """
    def __init__(self,
                 ref_molblock: str,
                 ref_partial_charges: np.array,
                 num_points: int = 400,
                 pharm_multi_vector: Optional[bool] = False,
                 probe_radius: float = 0.6,
                 xtb_opt: bool = True,
                 solvent: Optional[str] = 'water',
                 num_processes: int = 4):
        """
        Objective class that aligns fit molecules given a reference using ESP 3D similarity,
        and scoring with ESP+Pharmacophore combined 3D similarity.
        Constructor -- sets up the reference molecule and relevant representations.

        Arguments
        ---------
        ref_molblock : str molblock of the reference molecule.
        ref_partial_charges : np.array partial charges of reference molecule
        num_points : Optional[int] if `rep` is 'shape' or 'esp' this is necessary to generate the
            surface point cloud.
        pharm_multi_vector : Optional[bool] (default=False) whether to represent pharmacophores with
        mulitple vectors (if applicable) or a single, averaged vector. `None` does not generate
            pharmacophores.
        xtb_opt : bool to optimize conformers with xtb or just use EKTG embedding and MMFF opt.
        solvent : Optional[str] (default='water') solvent to use if optimizing with xtb.
            `None` is gas phase.
        num_processes : int (default = 1) Number of proccesses to use during xtb optimization
        """
        self.ref_partial_charges = None
        self.xtb_opt = xtb_opt
        self.solvent = solvent
        self.ref_molec = None
        self.probe_radius = probe_radius

        self.buffer = {} # stores past SMILES and their similarity scores

        self.num_points = num_points
        self.alpha = ALPHA(self.num_points)

        self.lam = 0.3

        self.pharm_multi_vector = pharm_multi_vector

        self.num_processes = num_processes

        self.sim_score_distr_with_resample = np.array([1.])
        self.sim_score_upper_bound = 1.

        ref_mol = Chem.MolFromMolBlock(ref_molblock, removeHs=False)
        self.ref_partial_charges = np.array(ref_partial_charges)

        # Get Molecule objects -- generate surf, esp, and pharms
        self.ref_molec = Molecule(mol=ref_mol,
                                  probe_radius=self.probe_radius,
                                  num_surf_points=self.num_points,
                                  partial_charges=self.ref_partial_charges,
                                  pharm_multi_vector=self.pharm_multi_vector)

        self.sim_score_distr_with_resample = self.resampling_surf_scores()
        self.sim_score_upper_bound = max(self.sim_score_distr_with_resample)

        self.buffer[Chem.MolToSmiles(self.ref_molec.mol)] = self.sim_score_upper_bound


    def resampling_surf_scores(self) -> np.ndarray:
        """
        Capture distribution of ESP surface + pharmacophore scores caused by resampling
        surface.

        Returns
        -------
        score_distr : np.ndarray (if not relevant)
        """
        num_repeats = 50

        esp_score_distr = np.empty(num_repeats)
        for i in range(num_repeats):
            molec = Molecule(mol=self.ref_molec.mol,
                             num_surf_points=self.ref_molec.num_surf_points,
                             probe_radius=self.ref_molec.probe_radius,
                             partial_charges=self.ref_molec.partial_charges)
            esp_score_distr[i] = get_overlap_esp_np(
                centers_1=self.ref_molec.surf_pos,
                centers_2=molec.surf_pos,
                charges_1=self.ref_molec.surf_esp,
                charges_2=molec.surf_esp,
                alpha=self.alpha,
                lam=self.lam
            )

        pharm_score = get_overlap_pharm_np(self.ref_molec.pharm_types, self.ref_molec.pharm_types,
                                         self.ref_molec.pharm_ancs, self.ref_molec.pharm_ancs,
                                         self.ref_molec.pharm_vecs, self.ref_molec.pharm_vecs,
                                         'tanimoto', extended_points=False, only_extended=False)

        return esp_score_distr + pharm_score


    def _score(self,
               fit_mol: Chem.Mol,
               fit_partial_charges: np.ndarray,
               trans_init: bool = False,
               use_jax: bool = False) -> float:
        """
        Align and score a fit molecule to the reference molecule.

        Arguments
        ---------
        fit_mol : Chem.Mol fit molecule to compare to reference.
        fit_partial_charges : np.ndarray of partial charges for each atom
        trans_init : bool (default = False) Whether to initially translate fit COM to each ref atom
            during alignment.
        use_jax : bool (default = False) uses torch if False for alignment

        Returns
        -------
        float : similarity score
        """
        fit_molec = Molecule(mol=fit_mol,
                             num_surf_points=self.num_points,
                             partial_charges=np.array(fit_partial_charges),
                             pharm_multi_vector=self.pharm_multi_vector,
                             probe_radius=self.ref_molec.probe_radius)

        molec_pair = MoleculePair(self.ref_molec, fit_molec, num_surf_points=self.num_points)

        molec_pair.align_with_esp(self.alpha,
                                  lam=self.lam,
                                  trans_init=trans_init,
                                  use_jax=use_jax)
        esp_score = molec_pair.sim_aligned_esp
        se3_transform_esp = molec_pair.transform_esp

        transformed_pharm_ancs = apply_SE3_transform_np(fit_molec.pharm_ancs, se3_transform_esp)
        transformed_pharm_vecs = apply_SO3_transform_np(fit_molec.pharm_vecs, se3_transform_esp)

        pharm_score = get_overlap_pharm_np(self.ref_molec.pharm_types, fit_molec.pharm_types,
                                         self.ref_molec.pharm_ancs, transformed_pharm_ancs,
                                         self.ref_molec.pharm_vecs, transformed_pharm_vecs,
                                         'tanimoto', extended_points=False, only_extended=False)
        return esp_score, pharm_score


    def score(self,
              smiles: str,
              num_conformers: int = 5,
              trans_init: bool = False,
              use_jax: bool = False) -> Tuple[float, float]:
        """
        Align and score a fit molecule to the reference molecule.

        Arguments
        ---------
        smiles : str SMILES string of molecule to compare to reference.
        num_conformers : int (default = 1) Max number of conformers to generate and score.
        trans_init : bool (default = False) Whether to initially translate fit COM to each ref atom
            during alignment.
        use_jax : bool (default = False) uses torch if False for alignment

        Returns
        -------
        Tuple
            float : ESP similarity score (max=1.)
            float : Pharm similarity score (max=1.)
        """
        fit_partial_charges = None
        if num_conformers == 1:
            fit_mol = embed_conformer_from_smiles(smiles, attempts=50, MMFF_optimize=True)
            charge = Chem.GetFormalCharge(fit_mol)

            if self.xtb_opt:
                fit_mol, _, fit_partial_charges = optimize_conformer_with_xtb(
                    fit_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    charge=charge,
                    temp_dir=TMPDIR
                )
            if fit_partial_charges is None:
                fit_partial_charges = charges_from_single_point_conformer_with_xtb(
                    fit_mol,
                    solvent=self.solvent,
                    num_cores=self.num_processes,
                    charge=charge,
                    temp_dir=TMPDIR
                )
            esp_score, pharm_score = self._score(
                fit_mol,
                fit_partial_charges=fit_partial_charges,
                trans_init=trans_init,
                use_jax=use_jax
            )
            return esp_score, pharm_score

        else:
            # if num_conformers is more than 1
            if self.xtb_opt:
                charge = Chem.GetFormalCharge(Chem.MolFromSmiles(smiles))
                fit_mols, _, fit_partial_charges = generate_opt_conformers_xtb(
                    smiles, charge=charge, MMFF_optimize=True, verbose=False, num_confs=num_conformers, temp_dir=TMPDIR
                )
            else:
                fit_mols = generate_opt_conformers(smiles, MMFF_optimize=True, verbose=False, num_confs=num_conformers)
                fit_partial_charges = []
                for m in fit_mols:
                    fit_partial_charges.append(charges_from_single_point_conformer_with_xtb(
                        conformer=m,
                        solvent=self.solvent,
                        num_cores=self.num_processes,
                        charge=charge,
                        temp_dir=TMPDIR
                    ))

            scores = []
            for i, fit_mol in enumerate(fit_mols):
                scores.append(self._score(fit_mol,
                                        fit_partial_charges=fit_partial_charges[i],
                                        trans_init=trans_init,
                                        use_jax=use_jax))
            total_scores = [sum(s) for s in scores]
            ind_max_score = np.argmax(np.array(total_scores))
            return scores[ind_max_score]


    def score_multiple(self,
                       smiles: List[str],
                       num_conformers: int = 5,
                       trans_init: bool = False,
                       use_jax: bool = False,
                       verbose=False) -> List[float]:
        """
        Aligns and scores multiple fit SMILES.

        Arguments
        ---------
        smiles : List[str] of SMILES strings of molecules to compare to reference.
        num_conformers : int (default = 1) Max number of conformers to generate and score.
        trans_init : bool (default = False) Whether to initially translate fit COM to each ref atom
            during alignment.
        use_jax : bool (default = False) uses torch if False for alignment

        Returns
        -------
        List[float] : similarity scores. Returns a value of -1 if Objective.score fails for any reason.
        """
        scores = []
        if verbose:
            pbar = tqdm(smiles, total=len(smiles))
        else:
            pbar = smiles
        for smi in pbar:
            try:
                # Canonicalize smiles
                smi = Chem.CanonSmiles(smi)
            except Exception:
                # if not a valid smiles skip
                scores.append(-1.)
                self.buffer[smi] = {'esp': None,
                                    'pharm': None}
                continue

            if smi in self.buffer:
                # skip if we've already computed it
                scores.append(self.buffer[smi]['esp'] + self.buffer[smi]['pharm'])
            else:
                try:
                    esp_score, pharm_score = self.score(
                        smi,
                        num_conformers=num_conformers,
                        trans_init=trans_init,
                        use_jax=use_jax
                    )
                    scores.append(esp_score + pharm_score)
                except Exception:
                    scores.append(-1.)
                    esp_score = None
                    pharm_score = None

            self.buffer[smi] = {'esp': esp_score,
                                'pharm': pharm_score}

        return scores



"""
Numpy Implementations for principal component alignment (pca).
Using the numpy version of quaternions_for_principal_component_alignment is faster than
the torch implementation (~2.5ms vs ~5ms).

Credit to Lewis J. Martin as this was adapted from
https://github.com/ljmartin/align/blob/main/0.2%20aligning%20principal%20moments%20of%20inertia.ipynb
"""
import numpy as np
from shepherd_score.alignment_utils.se3_np import get_SE3_transform_np, apply_SE3_transform_np

def compute_moment_of_inertia_np(points: np.ndarray) -> np.ndarray:
    """
    Computes the moment of inertia tensor for a set of points. Numpy implementation.
    A = x^2 + y^2 + z^2
    B = X^T X
    """
    points = points - np.mean(points, axis=0)
    A = np.sum(points**2)
    B = points.T @ points
    eye = np.eye(3)
    return (A * eye - B) / points.shape[0]


def compute_principal_moments_of_interia_np(points: np.ndarray) -> np.ndarray:
    """
    Calculate principal moment of inertia. Numpy implementation.
    """
    momint = compute_moment_of_inertia_np(points)
    eigvals, eigvecs = np.linalg.eigh(momint)
    indices = np.argsort(-eigvals) #sorting it returns the 'long' axis in index 0.
    # Return transposed which is more intuitive format
    return eigvecs[:, indices].T


def angle_between_vecs_np(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    """ Compute the angle in radians between two vectors. Numpy implementation. """
    v1_u = v1 / np.linalg.norm(v1)
    v2_u = v2 / np.linalg.norm(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def rotation_axis_np(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    """
    Calculate the vector about which to order to rotate `a` to align with `b` (cross product).
    Numpy implementation.
    """
    if np.allclose(v1, v2):
        return np.array([1, 0, 0])
    v3 = np.cross(v1, v2)
    return v3 / np.linalg.norm(v3)


def quaternion_from_axis_angle_np(axis: np.ndarray, angle: np.ndarray) -> np.ndarray:
    """
    Create a Quaternion from a rotation axis and an angle in radians.
    Numpy implementation.

    Parameters
    ----------
    axis : np.ndarray (3,)
        Axis to rotate about.
    angle: np.ndarray (1,)
        Angle in radians.

    Returns
    -------
    quaternion : np.ndarray (4,)
    """
    mag_sq = np.dot(axis, axis)
    if mag_sq == 0.0:
        raise ZeroDivisionError("Provided rotation axis has no length")
    # Ensure axis is in unit vector form
    if (abs(1.0 - mag_sq) > 1e-12):
        axis = axis / np.sqrt(mag_sq)
    theta = angle / 2.0
    r = np.cos(theta)
    i = axis * np.sin(theta)
    return np.array([r, i[0], i[1], i[2]])


def quaternion_mult_np(p: np.ndarray, q: np.ndarray) -> np.ndarray:
    """
    Multiplication of quaternions p and q. Numpy implementation.

    Reference: https://academicflight.com/articles/kinematics/rotation-formalisms/quaternions/

    General use case: The consecutive rotations of q_1 then q_2 is equivalent
    to q_3 = q_2*q_1. (order matters)

    Parameters
    ----------
    p : np.ndarray
        The first quaternion with shape (4,).
    q : np.ndarray
        The second quaternion with shape (4,).

    Returns
    -------
    np.ndarray
        The product of the two quaternions with shape (4,).
    """
    mat1 = np.array([[p[0], -p[1], -p[2], -p[3]],
                     [p[1],  p[0], -p[3],  p[2]],
                     [p[2],  p[3],  p[0], -p[1]],
                     [p[3], -p[2],  p[1],  p[0]]])
    pq = mat1 @ q
    return pq


def quaternions_for_principal_component_alignment_np(ref_points: np.ndarray,
                                                     fit_points: np.ndarray
                                                     ) -> np.ndarray:
    """
    Computes the 4 quaternions required for alignment of the fit mol along the
    principal components of the reference mol.
    NumPy implementation.

    The computed quaternions assumes that fit_points will be rotated after being centered at COM.

    Parameters
    ----------
    ref_points : np.ndarray (N, 3)
        Set of reference points that fit_points will be aligned to.
    fit_points : np.ndarray (M, 3)
        Set of points that will be aligned to ref_points.

    Returns
    -------
    quaternions : np.ndarray (4, 4)
        Set of four quaternions corresponding to the alignment of fit_points to ref_points in the
        four possible principal component combinations.
    """
    pmi_ref = compute_principal_moments_of_interia_np(ref_points)

    quaternions = np.zeros((4,4))
    for q_index in range(4):
        if q_index == 1:
            # flip orientation of longest axis
            pmi_ref[0] = -pmi_ref[0]
        elif q_index == 2:
            # unflip orientation of longest axis
            pmi_ref[0] = -pmi_ref[0]
            # flip orientation of 2nd longest axis
            pmi_ref[1] = -pmi_ref[1]
        elif q_index == 3:
            # flip orientation of both axes
            pmi_ref[0] = -pmi_ref[0]

        quat_order = np.zeros((2,4))
        # Initially center to COM
        fit_points_adjust = fit_points - np.mean(fit_points, axis=0)
        for ax_idx in range(2):
            pmi_fit = compute_principal_moments_of_interia_np(fit_points_adjust)
            # Angle between principal axis of fit mol and referencne mol
            angle = angle_between_vecs_np(pmi_fit[ax_idx], pmi_ref[ax_idx])
            # Axis that we are rotating about
            ax = rotation_axis_np(pmi_fit[ax_idx], pmi_ref[ax_idx])
            # Quaternion
            quat_order[ax_idx] = quaternion_from_axis_angle_np(ax, angle)
            # get SE(3) transformation matrix
            se3_params = np.concatenate((quat_order[ax_idx], np.zeros(3)))
            # get transformed matrix
            fit_points_adjust = apply_SE3_transform_np(fit_points_adjust, get_SE3_transform_np(se3_params))
        quaternions[q_index] = quaternion_mult_np(quat_order[1], quat_order[0])
    return quaternions



"""
Gaussian volume overlap scoring functions -- Shape-only (i.e., not color)
JAX VERSION (~ 6x faster than numpy)

Batched and non-batched functionalities

Reference math:
https://doi.org/10.1002/(SICI)1096-987X(19961115)17:14<1653::AID-JCC7>3.0.CO;2-K
https://doi.org/10.1021/j100011a016
"""
from jax import jit, Array
import jax.numpy as jnp
import jax


###################################################################################################
####### JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX JAX #######
###################################################################################################

@jit
def jax_cdist(X_1: Array,
              X_2: Array
              ) -> Array:
    """
    Jax implementation pairwise euclidian distances.

    Parameters
    ----------
    X_1 : Array (N, P)
    X_2 : Array (M, P)

    Returns
    -------
    Array (N, M)
        Distance matrix between X_1 and X_2.
    """
    distances = jnp.linalg.norm((X_1[:, None, :] - X_2[None, :, :]), axis=-1)
    return distances

@jit
def jax_sq_cdist(X_1: Array,
                 X_2: Array
                 ) -> Array:
    """
    Jax implementation pairwise SQUARED euclidian distances.

    Parameters
    ----------
    X_1 : Array (N, P)
    X_2 : Array (M, P)

    Returns
    -------
    Array (N, M)
        Distance matrix between X_1 and X_2, squared.
    """
    distances = jnp.sum(jnp.square((X_1[:, None, :] - X_2[None, :, :])), axis=-1)
    return distances


def VAB_2nd_order_jax(centers_1: Array, centers_2: Array, alpha: float) -> Array:
    """ 2nd order volume overlap of AB """
    R2 = jax_sq_cdist(centers_1, centers_2)

    VAB_2nd_order = jnp.sum(jnp.pi**(1.5) * jnp.exp(-(alpha / 2) * R2) / ((2*alpha)**(1.5)))
    return VAB_2nd_order


def shape_tanimoto_jax(centers_1: Array, centers_2: Array, alpha: float) -> Array:
    """ Compute Tanimoto shape similarity """
    VAA = VAB_2nd_order_jax(centers_1, centers_1, alpha)
    VBB = VAB_2nd_order_jax(centers_2, centers_2, alpha)
    VAB = VAB_2nd_order_jax(centers_1, centers_2, alpha)
    return VAB / (VAA + VBB - VAB)

@jit
def get_overlap_jax(centers_1: Array,
                    centers_2: Array,
                    alpha: float = 0.81
                    ) -> Array:
    """ Compute ROCS Gaussian volume overlap using jitted jax function. """
    tanimoto = shape_tanimoto_jax(centers_1, centers_2, alpha)
    return tanimoto


@jit
def _mask_prod_jax(mask_1: Array, mask_2: Array):
    return mask_1[:, None] * mask_2[None, :]


def VAB_2nd_order_jax_mask(centers_1: Array, centers_2: Array,
                            mask_1: Array, mask_2: Array,
                            alpha: float) -> Array:
    """ 2nd order volume overlap of AB """
    R2 = jax_sq_cdist(centers_1, centers_2)
    M2 = _mask_prod_jax(mask_1, mask_2)

    VAB_2nd_order = jnp.sum(M2 * jnp.pi**(1.5) * jnp.exp(-(alpha / 2) * R2) / ((2*alpha)**(1.5)))
    return VAB_2nd_order


def shape_tanimoto_jax_mask(centers_1: Array, centers_2: Array,
                            mask_1: Array, mask_2: Array,
                            alpha: float) -> Array:
    """ Compute Tanimoto shape similarity """
    VAA = VAB_2nd_order_jax_mask(centers_1, centers_1, mask_1, mask_1, alpha)
    VBB = VAB_2nd_order_jax_mask(centers_2, centers_2, mask_2, mask_2, alpha)
    VAB = VAB_2nd_order_jax_mask(centers_1, centers_2, mask_1, mask_2, alpha)
    return VAB / (VAA + VBB - VAB)

@jit
def get_overlap_jax_mask(centers_1: Array,
                         centers_2: Array,
                         mask_1: Array,
                         mask_2: Array,
                         alpha: float = 0.81
                         ) -> Array:
    """ Compute ROCS Gaussian volume overlap using jitted jax function. """
    tanimoto = shape_tanimoto_jax_mask(centers_1, centers_2, mask_1, mask_2, alpha)
    return tanimoto


def _VAB_2nd_order_cosine_jax(centers_1: Array,
                              centers_2: Array,
                              vectors_1: Array,
                              vectors_2: Array,
                              alpha: float,
                              allow_antiparallel: bool,
                              ) -> Array:
    """
    2nd order volume overlap of AB weighted by cosine similarity (JAX version) - implementation part.
    """
    R2 = jax_sq_cdist(centers_1, centers_2)  # (N1, N2)
    term_common = (jnp.pi**1.5) / ((2 * alpha)**1.5)

    # Normalize vectors
    vec1_norm = vectors_1 / jnp.linalg.norm(vectors_1, axis=-1, keepdims=True)
    vec2_norm = vectors_2 / jnp.linalg.norm(vectors_2, axis=-1, keepdims=True)

    # Cosine similarity: (N1, N2)
    V2_sim = jnp.dot(vec1_norm, vec2_norm.T)

    V2_sim = jax.lax.cond(
        allow_antiparallel,
        lambda x: jnp.abs(x),    # True branch
        lambda x: jnp.clip(x, 0., 1.),  # False branch
        V2_sim
    )
    V2_weighted = (V2_sim + 2.) / 3.

    VAB_second_order = jnp.sum(term_common *
                               V2_weighted *  # REMOVED .T : V2_weighted is (N1,N2), R2 is (N1,N2)
                               jnp.exp(-(alpha / 2) * R2))
    return VAB_second_order

VAB_2nd_order_cosine_jax = jit(_VAB_2nd_order_cosine_jax, static_argnames=["allow_antiparallel"])


def _VAB_2nd_order_cosine_jax_mask(centers_1: Array,
                                   centers_2: Array,
                                   vectors_1: Array,
                                   vectors_2: Array,
                                   mask_1: Array,
                                   mask_2: Array,
                                   alpha: float,
                                   allow_antiparallel: bool,
                                   ) -> Array:
    """
    2nd order volume overlap of AB weighted by cosine similarity (JAX version) - implementation part.
    Vectors are assumed to be normalized.
    """
    R2 = jax_sq_cdist(centers_1, centers_2)  # (N1, N2)
    M2 = _mask_prod_jax(mask_1, mask_2)
    term_common = (jnp.pi**1.5) / ((2 * alpha)**1.5)

    # Cosine similarity: (N1, N2)
    V2_sim = jnp.dot(vectors_1, vectors_2.T)

    V2_sim = jax.lax.cond(
        allow_antiparallel,
        lambda x: jnp.abs(x),    # True branch
        lambda x: jnp.clip(x, 0., 1.),  # False branch
        V2_sim
    )
    V2_weighted = (V2_sim + 2.) / 3.

    VAB_second_order = jnp.sum(term_common *
                                 M2 *
                                 V2_weighted *
                                 jnp.exp(-(alpha / 2) * R2))
    return VAB_second_order

VAB_2nd_order_cosine_jax_mask = jit(_VAB_2nd_order_cosine_jax_mask, static_argnames=["allow_antiparallel"])



"""
Unit tests for the electrostatic_scoring module (PyTorch and JAX implementations).
Assumes the NumPy implementation is ground truth.

Particularly for the batching functionality that complicates the PyTorch implementation.
"""

import pytest
import torch
import numpy as np
from shepherd_score.score import electrostatic_scoring as es
from shepherd_score.score import electrostatic_scoring_np as es_np
from shepherd_score.score.constants import LAM_SCALING # For default lam values
from .utils import _configure_jax_platform

# Attempt to import JAX and related modules
JAX_AVAILABLE = False
jax = None
jnp = None
es_jax = None

try:
    # Configure JAX platform before import to avoid GPU initialization errors
    _gpu_detected = _configure_jax_platform()

    import jax
    import jax.numpy as jnp
    from shepherd_score.score import electrostatic_scoring_jax as es_jax

    JAX_AVAILABLE = True
except ImportError:
    # JAX not available - all tests in JAX test classes will be skipped
    pass


# Helper function to generate random electrostatic data
def _generate_electrostatic_data(batch_size=None,
                                 num_centers1=10, num_centers2=12,
                                 num_atoms_w_H1=15, num_atoms_w_H2=18,
                                 num_surf_points1=20, num_surf_points2=22,
                                 dim=3, fixed_seed=None):
    if fixed_seed is not None:
        np.random.seed(fixed_seed)

    def _generate_coords(count, bs, d):
        if bs:
            return np.random.rand(bs, count, d).astype(np.float32)
        return np.random.rand(count, d).astype(np.float32)

    def _generate_charges_or_radii(count, bs):
        if bs:
            return np.random.rand(bs, count).astype(np.float32)
        return np.random.rand(count).astype(np.float32)

    # NumPy arrays
    centers_1_np = _generate_coords(num_centers1, batch_size, dim)
    centers_2_np = _generate_coords(num_centers2, batch_size, dim)
    charges_overlap_1_np = _generate_charges_or_radii(num_centers1, batch_size)
    charges_overlap_2_np = _generate_charges_or_radii(num_centers2, batch_size)
    centers_w_H_1_np = _generate_coords(num_atoms_w_H1, batch_size, dim)
    centers_w_H_2_np = _generate_coords(num_atoms_w_H2, batch_size, dim)
    partial_charges_1_np = _generate_charges_or_radii(num_atoms_w_H1, batch_size)
    partial_charges_2_np = _generate_charges_or_radii(num_atoms_w_H2, batch_size)
    points_1_np = _generate_coords(num_surf_points1, batch_size, dim)
    points_2_np = _generate_coords(num_surf_points2, batch_size, dim)
    point_charges_1_np = _generate_charges_or_radii(num_surf_points1, batch_size)
    point_charges_2_np = _generate_charges_or_radii(num_surf_points2, batch_size)
    radii_1_np = _generate_charges_or_radii(num_atoms_w_H1, batch_size)
    radii_2_np = _generate_charges_or_radii(num_atoms_w_H2, batch_size)

    # PyTorch tensors
    centers_1_torch = torch.from_numpy(centers_1_np)
    centers_2_torch = torch.from_numpy(centers_2_np)
    charges_overlap_1_torch = torch.from_numpy(charges_overlap_1_np)
    charges_overlap_2_torch = torch.from_numpy(charges_overlap_2_np)
    centers_w_H_1_torch = torch.from_numpy(centers_w_H_1_np)
    centers_w_H_2_torch = torch.from_numpy(centers_w_H_2_np)
    partial_charges_1_torch = torch.from_numpy(partial_charges_1_np)
    partial_charges_2_torch = torch.from_numpy(partial_charges_2_np)
    points_1_torch = torch.from_numpy(points_1_np)
    points_2_torch = torch.from_numpy(points_2_np)
    point_charges_1_torch = torch.from_numpy(point_charges_1_np)
    point_charges_2_torch = torch.from_numpy(point_charges_2_np)
    radii_1_torch = torch.from_numpy(radii_1_np)
    radii_2_torch = torch.from_numpy(radii_2_np)

    # JAX arrays
    if JAX_AVAILABLE:
        centers_1_jnp = jnp.array(centers_1_np)
        centers_2_jnp = jnp.array(centers_2_np)
        charges_overlap_1_jnp = jnp.array(charges_overlap_1_np)
        charges_overlap_2_jnp = jnp.array(charges_overlap_2_np)
        centers_w_H_1_jnp = jnp.array(centers_w_H_1_np)
        centers_w_H_2_jnp = jnp.array(centers_w_H_2_np)
        partial_charges_1_jnp = jnp.array(partial_charges_1_np)
        partial_charges_2_jnp = jnp.array(partial_charges_2_np)
        points_1_jnp = jnp.array(points_1_np)
        points_2_jnp = jnp.array(points_2_np)
        point_charges_1_jnp = jnp.array(point_charges_1_np)
        point_charges_2_jnp = jnp.array(point_charges_2_np)
        radii_1_jnp = jnp.array(radii_1_np)
        radii_2_jnp = jnp.array(radii_2_np)
    else:
        centers_1_jnp = None
        centers_2_jnp = None
        charges_overlap_1_jnp = None
        charges_overlap_2_jnp = None
        centers_w_H_1_jnp = None
        centers_w_H_2_jnp = None
        partial_charges_1_jnp = None
        partial_charges_2_jnp = None
        points_1_jnp = None
        points_2_jnp = None
        point_charges_1_jnp = None
        point_charges_2_jnp = None
        radii_1_jnp = None
        radii_2_jnp = None

    return (centers_1_np, centers_2_np, charges_overlap_1_np, charges_overlap_2_np,
            centers_w_H_1_np, centers_w_H_2_np, partial_charges_1_np, partial_charges_2_np,
            points_1_np, points_2_np, point_charges_1_np, point_charges_2_np,
            radii_1_np, radii_2_np,
            centers_1_torch, centers_2_torch, charges_overlap_1_torch, charges_overlap_2_torch,
            centers_w_H_1_torch, centers_w_H_2_torch, partial_charges_1_torch, partial_charges_2_torch,
            points_1_torch, points_2_torch, point_charges_1_torch, point_charges_2_torch,
            radii_1_torch, radii_2_torch,
            centers_1_jnp, centers_2_jnp, charges_overlap_1_jnp, charges_overlap_2_jnp,
            centers_w_H_1_jnp, centers_w_H_2_jnp, partial_charges_1_jnp, partial_charges_2_jnp,
            points_1_jnp, points_2_jnp, point_charges_1_jnp, point_charges_2_jnp,
            radii_1_jnp, radii_2_jnp)


# Fixed parameters for tests
ALPHA_ESP = 0.81
LAM_ESP_OVERLAP = 0.3 * LAM_SCALING # Default for get_overlap_esp
LAM_ESP_COMPARISON = 0.001 # Default for _esp_comparison
PROBE_RADIUS = 1.0
ESP_WEIGHT = 0.5

class TestElectrostaticScoringTorch:
    def test_VAB_2nd_order_esp_single_instance(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        (c1_np, c2_np, ch1_np, ch2_np, _, _, _, _, _, _, _, _, _, _,
         c1_torch, c2_torch, ch1_torch, ch2_torch, *_) = _generate_electrostatic_data(fixed_seed=1)

        expected = es_np.VAB_2nd_order_esp_np(c1_np, c2_np, ch1_np.reshape(-1,1), ch2_np.reshape(-1,1), alpha, lam)
        actual = es.VAB_2nd_order_esp(c1_torch, c2_torch, ch1_torch.reshape(-1,1), ch2_torch.reshape(-1,1), alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_VAB_2nd_order_esp_batched(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 3
        (c1_np_b, c2_np_b, ch1_np_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         c1_torch_b, c2_torch_b, ch1_torch_b, ch2_torch_b, *_) = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=2)

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np.VAB_2nd_order_esp_np(c1_np_b[i], c2_np_b[i],
                                                           ch1_np_b[i].reshape(-1,1), ch2_np_b[i].reshape(-1,1),
                                                           alpha, lam))
        expected = np.array(expected_batch)
        actual = es.VAB_2nd_order_esp(c1_torch_b, c2_torch_b, ch1_torch_b.unsqueeze(-1), ch2_torch_b.unsqueeze(-1), alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_shape_tanimoto_esp_single_instance(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        (c1_np, c2_np, ch1_np, ch2_np, _, _, _, _, _, _, _, _, _, _,
         c1_torch, c2_torch, ch1_torch, ch2_torch, *_) = _generate_electrostatic_data(fixed_seed=3)

        expected = es_np.shape_tanimoto_esp_np(c1_np, c2_np, ch1_np.reshape(-1,1), ch2_np.reshape(-1,1), alpha, lam)
        actual = es.shape_tanimoto_esp(c1_torch, c2_torch, ch1_torch.reshape(-1,1), ch2_torch.reshape(-1,1), alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_shape_tanimoto_esp_batched(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 3
        (c1_np_b, c2_np_b, ch1_np_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         c1_torch_b, c2_torch_b, ch1_torch_b, ch2_torch_b, *_) = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=4)

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np.shape_tanimoto_esp_np(c1_np_b[i], c2_np_b[i],
                                                            ch1_np_b[i].reshape(-1,1), ch2_np_b[i].reshape(-1,1),
                                                            alpha, lam))
        expected = np.array(expected_batch)
        actual = es.shape_tanimoto_esp(c1_torch_b, c2_torch_b, ch1_torch_b.unsqueeze(-1), ch2_torch_b.unsqueeze(-1), alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_get_overlap_esp_single_instance(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        (c1_np, c2_np, ch1_np, ch2_np, _, _, _, _, _, _, _, _, _, _,
         c1_torch, c2_torch, ch1_torch, ch2_torch, *_) = _generate_electrostatic_data(fixed_seed=5)

        expected = es_np.get_overlap_esp_np(c1_np, c2_np, ch1_np, ch2_np, alpha, lam)
        actual = es.get_overlap_esp(c1_torch, c2_torch, ch1_torch, ch2_torch, alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_get_overlap_esp_batched(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 3
        (c1_np_b, c2_np_b, ch1_np_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         c1_torch_b, c2_torch_b, ch1_torch_b, ch2_torch_b, *_) = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=6)

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np.get_overlap_esp_np(c1_np_b[i], c2_np_b[i], ch1_np_b[i], ch2_np_b[i], alpha, lam))
        expected = np.array(expected_batch)
        actual = es.get_overlap_esp(c1_torch_b, c2_torch_b, ch1_torch_b, ch2_torch_b, alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_get_overlap_esp_broadcast_c1_ch1(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 3
        num_centers1 = 10
        num_centers2 = 12
        data_full_batch = _generate_electrostatic_data(batch_size=batch_size, num_centers1=num_centers1, num_centers2=num_centers2, fixed_seed=7)
        (c1_np_orig_b, c2_np_b, ch1_np_orig_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         c1_torch_orig_b, c2_torch_b, ch1_torch_orig_b, ch2_torch_b, *_) = data_full_batch

        c1_np_single = c1_np_orig_b[0]
        ch1_np_single = ch1_np_orig_b[0]
        c1_torch_single = c1_torch_orig_b[0]
        ch1_torch_single = ch1_torch_orig_b[0]

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np.get_overlap_esp_np(c1_np_single, c2_np_b[i], ch1_np_single, ch2_np_b[i], alpha, lam))
        expected = np.array(expected_batch)
        actual = es.get_overlap_esp(c1_torch_single, c2_torch_b, ch1_torch_single, ch2_torch_b, alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_get_overlap_esp_broadcast_c2_ch2(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 3
        num_centers1 = 10
        num_centers2 = 12
        data_full_batch = _generate_electrostatic_data(batch_size=batch_size, num_centers1=num_centers1, num_centers2=num_centers2, fixed_seed=8)
        (c1_np_b, c2_np_orig_b, ch1_np_b, ch2_np_orig_b, _, _, _, _, _, _, _, _, _, _,
         c1_torch_b, c2_torch_orig_b, ch1_torch_b, ch2_torch_orig_b, *_) = data_full_batch

        c2_np_single = c2_np_orig_b[0]
        ch2_np_single = ch2_np_orig_b[0]
        c2_torch_single = c2_torch_orig_b[0]
        ch2_torch_single = ch2_torch_orig_b[0]

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np.get_overlap_esp_np(c1_np_b[i], c2_np_single, ch1_np_b[i], ch2_np_single, alpha, lam))
        expected = np.array(expected_batch)
        actual = es.get_overlap_esp(c1_torch_b, c2_torch_single, ch1_torch_b, ch2_torch_single, alpha, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_esp_comparison_single_instance_all(self):
        """Test _esp_comparison with all inputs as single instances."""
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        data = _generate_electrostatic_data(fixed_seed=9)

        points_1_np = data[8]
        point_charges_1_np = data[10]
        centers_w_H_2_np = data[5]
        partial_charges_2_np = data[7]
        radii_2_np = data[13]

        points_1_torch = data[22]
        point_charges_1_torch = data[24]
        centers_w_H_2_torch = data[19]
        partial_charges_2_torch = data[21]
        radii_2_torch = data[27]

        expected = es_np._esp_comparison_np(points_1_np, centers_w_H_2_np, partial_charges_2_np,
                                            point_charges_1_np, radii_2_np, probe_radius, lam)
        actual = es._esp_comparison(points_1_torch, centers_w_H_2_torch, partial_charges_2_torch,
                                    point_charges_1_torch, radii_2_torch, probe_radius, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_esp_comparison_points1_single_mol2_batched(self):
        """Test _esp_comparison with points_1 single, and molecule 2 data batched."""
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        batch_size = 2
        num_surf_points1 = 20
        num_atoms_w_H2 = 18

        data = _generate_electrostatic_data(batch_size=batch_size,
                                            num_surf_points1=num_surf_points1,
                                            num_atoms_w_H2=num_atoms_w_H2,
                                            fixed_seed=10)
        points_1_np = data[8][0]
        point_charges_1_np = data[10][0]
        points_1_torch = torch.from_numpy(points_1_np)
        point_charges_1_torch = torch.from_numpy(point_charges_1_np)

        centers_w_H_2_np_b = data[5]
        partial_charges_2_np_b = data[7]
        radii_2_np_b = data[13]
        centers_w_H_2_torch_b = data[19]
        partial_charges_2_torch_b = data[21]
        radii_2_torch_b = data[27]

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np._esp_comparison_np(points_1_np, centers_w_H_2_np_b[i],
                                                           partial_charges_2_np_b[i], point_charges_1_np,
                                                           radii_2_np_b[i], probe_radius, lam))
        expected = np.array(expected_batch)
        actual = es._esp_comparison(points_1_torch, centers_w_H_2_torch_b, partial_charges_2_torch_b,
                                    point_charges_1_torch, radii_2_torch_b, probe_radius, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_esp_comparison_points1_batched_mol2_single(self):
        """Test _esp_comparison with points_1 batched, and molecule 2 data single (broadcast)."""
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        batch_size = 2
        num_surf_points1 = 20
        num_atoms_w_H2 = 18

        data = _generate_electrostatic_data(batch_size=batch_size,
                                            num_surf_points1=num_surf_points1,
                                            num_atoms_w_H2=num_atoms_w_H2,
                                            fixed_seed=11)
        points_1_np_b = data[8]
        point_charges_1_np_b = data[10]
        points_1_torch_b = data[22]
        point_charges_1_torch_b = data[24]

        centers_w_H_2_np_s = data[5][0]
        partial_charges_2_np_s = data[7][0]
        radii_2_np_s = data[13][0]
        centers_w_H_2_torch_s = torch.from_numpy(centers_w_H_2_np_s)
        partial_charges_2_torch_s = torch.from_numpy(partial_charges_2_np_s)
        radii_2_torch_s = torch.from_numpy(radii_2_np_s)

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np._esp_comparison_np(points_1_np_b[i], centers_w_H_2_np_s,
                                                           partial_charges_2_np_s, point_charges_1_np_b[i],
                                                           radii_2_np_s, probe_radius, lam))
        expected = np.array(expected_batch)
        actual = es._esp_comparison(points_1_torch_b, centers_w_H_2_torch_s, partial_charges_2_torch_s,
                                    point_charges_1_torch_b, radii_2_torch_s, probe_radius, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_esp_comparison_all_batched(self):
        """Test _esp_comparison with all inputs batched."""
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        batch_size = 2
        data = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=12)

        points_1_np_b = data[8]
        point_charges_1_np_b = data[10]
        centers_w_H_2_np_b = data[5]
        partial_charges_2_np_b = data[7]
        radii_2_np_b = data[13]

        points_1_torch_b = data[22]
        point_charges_1_torch_b = data[24]
        centers_w_H_2_torch_b = data[19]
        partial_charges_2_torch_b = data[21]
        radii_2_torch_b = data[27]

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(es_np._esp_comparison_np(points_1_np_b[i], centers_w_H_2_np_b[i],
                                                           partial_charges_2_np_b[i], point_charges_1_np_b[i],
                                                           radii_2_np_b[i], probe_radius, lam))
        expected = np.array(expected_batch)
        actual = es._esp_comparison(points_1_torch_b, centers_w_H_2_torch_b, partial_charges_2_torch_b,
                                    point_charges_1_torch_b, radii_2_torch_b, probe_radius, lam)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_esp_combo_score_single_instance(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        esp_weight = ESP_WEIGHT

        (centers_overlap_1_np, centers_overlap_2_np, _, _,
        centers_w_H_1_np, centers_w_H_2_np, partial_charges_1_np, partial_charges_2_np,
        points_1_np, points_2_np, point_charges_1_np, point_charges_2_np,
        radii_1_np, radii_2_np,
        centers_overlap_1_torch, centers_overlap_2_torch, _, _,
        centers_w_H_1_torch, centers_w_H_2_torch, partial_charges_1_torch, partial_charges_2_torch,
        points_1_torch, points_2_torch, point_charges_1_torch, point_charges_2_torch,
        radii_1_torch, radii_2_torch, *_) = _generate_electrostatic_data(fixed_seed=13, num_centers1=5, num_centers2=6)

        expected = es_np.esp_combo_score_np(centers_w_H_1_np, centers_w_H_2_np,
                                              centers_overlap_1_np, centers_overlap_2_np,
                                              points_1_np, points_2_np,
                                              partial_charges_1_np, partial_charges_2_np,
                                              point_charges_1_np, point_charges_2_np,
                                              radii_1_np, radii_2_np,
                                              alpha, lam, probe_radius, esp_weight)
        actual = es.esp_combo_score(centers_w_H_1_torch, centers_w_H_2_torch,
                                    centers_overlap_1_torch, centers_overlap_2_torch,
                                    points_1_torch, points_2_torch,
                                    partial_charges_1_torch, partial_charges_2_torch,
                                    point_charges_1_torch, point_charges_2_torch,
                                    radii_1_torch, radii_2_torch,
                                    alpha, lam, probe_radius, esp_weight)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

    def test_esp_combo_score_batched(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        esp_weight = ESP_WEIGHT
        batch_size = 2

        data_b = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=14, num_centers1=5, num_centers2=6)
        (centers_overlap_1_np_b, centers_overlap_2_np_b, _, _,
        centers_w_H_1_np_b, centers_w_H_2_np_b, partial_charges_1_np_b, partial_charges_2_np_b,
        points_1_np_b, points_2_np_b, point_charges_1_np_b, point_charges_2_np_b,
        radii_1_np_b, radii_2_np_b,
        centers_overlap_1_torch_b, centers_overlap_2_torch_b, _, _,
        centers_w_H_1_torch_b, centers_w_H_2_torch_b, partial_charges_1_torch_b, partial_charges_2_torch_b,
        points_1_torch_b, points_2_torch_b, point_charges_1_torch_b, point_charges_2_torch_b,
        radii_1_torch_b, radii_2_torch_b, *_) = data_b

        expected_batch = []
        for i in range(batch_size):
            expected_batch.append(
                es_np.esp_combo_score_np(centers_w_H_1_np_b[i], centers_w_H_2_np_b[i],
                                         centers_overlap_1_np_b[i], centers_overlap_2_np_b[i],
                                         points_1_np_b[i], points_2_np_b[i],
                                         partial_charges_1_np_b[i], partial_charges_2_np_b[i],
                                         point_charges_1_np_b[i], point_charges_2_np_b[i],
                                         radii_1_np_b[i], radii_2_np_b[i],
                                         alpha, lam, probe_radius, esp_weight)
            )
        expected = np.array(expected_batch)
        actual = es.esp_combo_score(centers_w_H_1_torch_b, centers_w_H_2_torch_b,
                                    centers_overlap_1_torch_b, centers_overlap_2_torch_b,
                                    points_1_torch_b, points_2_torch_b,
                                    partial_charges_1_torch_b, partial_charges_2_torch_b,
                                    point_charges_1_torch_b, point_charges_2_torch_b,
                                    radii_1_torch_b, radii_2_torch_b,
                                    alpha, lam, probe_radius, esp_weight)
        np.testing.assert_allclose(actual.numpy(), expected, rtol=1e-5)

# Placeholder for JAX tests
@pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX is not installed")
@pytest.mark.jax
class TestElectrostaticScoringJAX:
    def test_VAB_2nd_order_esp_jax(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        (c1_np, c2_np, ch1_np, ch2_np, _, _, _, _, _, _, _, _, _, _,
         _, _, _, _, _, _, _, _, _, _, _, _, _, _,
         c1_jnp, c2_jnp, ch1_jnp, ch2_jnp, *_) = _generate_electrostatic_data(fixed_seed=101)

        expected_np = es_np.VAB_2nd_order_esp_np(c1_np, c2_np, ch1_np.reshape(-1,1), ch2_np.reshape(-1,1), alpha, lam)
        # JAX charges are (N,) and need to be (N,1) for cdist as in numpy version
        actual_jax = es_jax.VAB_2nd_order_esp_jax(c1_jnp, c2_jnp, ch1_jnp.reshape(-1,1), ch2_jnp.reshape(-1,1), alpha, lam)
        np.testing.assert_allclose(np.array(actual_jax), expected_np, rtol=1e-5)

    def test_shape_tanimoto_esp_jax(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        (c1_np, c2_np, ch1_np, ch2_np, _, _, _, _, _, _, _, _, _, _,
         _, _, _, _, _, _, _, _, _, _, _, _, _, _,
         c1_jnp, c2_jnp, ch1_jnp, ch2_jnp, *_) = _generate_electrostatic_data(fixed_seed=102)

        expected_np = es_np.shape_tanimoto_esp_np(c1_np, c2_np, ch1_np.reshape(-1,1), ch2_np.reshape(-1,1), alpha, lam)
        # JAX charges are (N,) and need to be (N,1) for cdist as in numpy version for underlying VAB calls
        actual_jax = es_jax.shape_tanimoto_esp_jax(c1_jnp, c2_jnp, ch1_jnp.reshape(-1,1), ch2_jnp.reshape(-1,1), alpha, lam)
        np.testing.assert_allclose(np.array(actual_jax), expected_np, rtol=1e-5)

    def test_get_overlap_esp_jax(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        (c1_np, c2_np, ch1_np, ch2_np, _, _, _, _, _, _, _, _, _, _,
         _, _, _, _, _, _, _, _, _, _, _, _, _, _,
         c1_jnp, c2_jnp, ch1_jnp, ch2_jnp, *_) = _generate_electrostatic_data(fixed_seed=103)

        expected_np = es_np.get_overlap_esp_np(c1_np, c2_np, ch1_np, ch2_np, alpha, lam)
        # get_overlap_esp_jax is expected to handle reshaping of charges if necessary, like PyTorch version
        actual_jax = es_jax.get_overlap_esp_jax(c1_jnp, c2_jnp, ch1_jnp, ch2_jnp, alpha, lam)
        np.testing.assert_allclose(np.array(actual_jax), expected_np, rtol=1e-5)

    def test_esp_comparison_jax(self):
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        data = _generate_electrostatic_data(fixed_seed=104)

        points_1_np = data[8]
        point_charges_1_np = data[10]
        centers_w_H_2_np = data[5]
        partial_charges_2_np = data[7]
        radii_2_np = data[13]

        points_1_jnp = data[36]
        point_charges_1_jnp = data[38]
        centers_w_H_2_jnp = data[33]
        partial_charges_2_jnp = data[35]
        radii_2_jnp = data[41]

        expected_np = es_np._esp_comparison_np(points_1_np, centers_w_H_2_np, partial_charges_2_np,
                                               point_charges_1_np, radii_2_np, probe_radius, lam)
        actual_jax = es_jax._esp_comparison_jax(points_1_jnp, centers_w_H_2_jnp, partial_charges_2_jnp,
                                              point_charges_1_jnp, radii_2_jnp, probe_radius, lam)
        np.testing.assert_allclose(np.array(actual_jax), expected_np, rtol=1e-5)

    def test_esp_combo_score_jax(self):
        alpha = ALPHA_ESP
        lam_comp = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        esp_weight = ESP_WEIGHT

        data = _generate_electrostatic_data(fixed_seed=105, num_centers1=5, num_centers2=6)
        (centers_overlap_1_np, centers_overlap_2_np, _, _,
        centers_w_H_1_np, centers_w_H_2_np, partial_charges_1_np, partial_charges_2_np,
        points_1_np, points_2_np, point_charges_1_np, point_charges_2_np,
        radii_1_np, radii_2_np,
        _,_,_,_,_,_,_,_,_,_,_,_,_,_,
        centers_overlap_1_jnp, centers_overlap_2_jnp, _, _,
        centers_w_H_1_jnp, centers_w_H_2_jnp, partial_charges_1_jnp, partial_charges_2_jnp,
        points_1_jnp, points_2_jnp, point_charges_1_jnp, point_charges_2_jnp,
        radii_1_jnp, radii_2_jnp) = data

        expected_np = es_np.esp_combo_score_np(centers_w_H_1_np, centers_w_H_2_np,
                                             centers_overlap_1_np, centers_overlap_2_np,
                                             points_1_np, points_2_np,
                                             partial_charges_1_np, partial_charges_2_np,
                                             point_charges_1_np, point_charges_2_np,
                                             radii_1_np, radii_2_np,
                                             alpha, lam_comp, probe_radius, esp_weight)
        actual_jax = es_jax.esp_combo_score_jax(centers_w_H_1_jnp, centers_w_H_2_jnp,
                                           centers_overlap_1_jnp, centers_overlap_2_jnp,
                                           points_1_jnp, points_2_jnp,
                                           partial_charges_1_jnp, partial_charges_2_jnp,
                                           point_charges_1_jnp, point_charges_2_jnp,
                                           radii_1_jnp, radii_2_jnp,
                                           alpha, lam_comp, probe_radius, esp_weight)
        np.testing.assert_allclose(np.array(actual_jax), expected_np, rtol=1e-5)

@pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX is not installed")
@pytest.mark.jax
class TestElectrostaticScoringJAXBatched:
    def test_VAB_2nd_order_esp_jax_vmap_batch(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 2
        (c1_np_b, c2_np_b, ch1_np_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         _, _, _, _, _, _, _, _, _, _, _, _, _, _,
         c1_jnp_b, c2_jnp_b, ch1_jnp_b, ch2_jnp_b, *_) = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=201)

        # Vmap: centers and charges are batched (axis 0), alpha and lam are scalar (None)
        # Charges need to be (B, N, 1) for the cdist within the JAX function if it mirrors NumPy version
        vmapped_VAB_jax = jax.vmap(es_jax.VAB_2nd_order_esp_jax,
                                   in_axes=(0, 0, 0, 0, None, None)) # c1, c2, ch1, ch2, alpha, lam

        expected_np_b = []
        for i in range(batch_size):
            expected_np_b.append(es_np.VAB_2nd_order_esp_np(c1_np_b[i], c2_np_b[i],
                                                            ch1_np_b[i].reshape(-1,1), ch2_np_b[i].reshape(-1,1),
                                                            alpha, lam))
        expected_np_array = np.array(expected_np_b)

        actual_jax_b = vmapped_VAB_jax(c1_jnp_b, c2_jnp_b, ch1_jnp_b.reshape(batch_size,-1,1), ch2_jnp_b.reshape(batch_size,-1,1), alpha, lam)
        np.testing.assert_allclose(np.array(actual_jax_b), expected_np_array, rtol=1e-5)

    def test_shape_tanimoto_esp_jax_vmap_batch(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 2
        (c1_np_b, c2_np_b, ch1_np_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         _, _, _, _, _, _, _, _, _, _, _, _, _, _,
         c1_jnp_b, c2_jnp_b, ch1_jnp_b, ch2_jnp_b, *_) = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=202)

        vmapped_tanimoto_jax = jax.vmap(es_jax.shape_tanimoto_esp_jax,
                                        in_axes=(0, 0, 0, 0, None, None)) # c1, c2, ch1, ch2, alpha, lam

        expected_np_b = []
        for i in range(batch_size):
            expected_np_b.append(es_np.shape_tanimoto_esp_np(c1_np_b[i], c2_np_b[i],
                                                             ch1_np_b[i].reshape(-1,1), ch2_np_b[i].reshape(-1,1),
                                                             alpha, lam))
        expected_np_array = np.array(expected_np_b)

        actual_jax_b = vmapped_tanimoto_jax(c1_jnp_b, c2_jnp_b, ch1_jnp_b.reshape(batch_size,-1,1), ch2_jnp_b.reshape(batch_size,-1,1), alpha, lam)
        np.testing.assert_allclose(np.array(actual_jax_b), expected_np_array, rtol=1e-5)

    def test_get_overlap_esp_jax_vmap_batch(self):
        alpha = ALPHA_ESP
        lam = LAM_ESP_OVERLAP
        batch_size = 2
        (c1_np_b, c2_np_b, ch1_np_b, ch2_np_b, _, _, _, _, _, _, _, _, _, _,
         _, _, _, _, _, _, _, _, _, _, _, _, _, _,
         c1_jnp_b, c2_jnp_b, ch1_jnp_b, ch2_jnp_b, *_) = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=203)

        vmapped_get_overlap_jax = jax.vmap(es_jax.get_overlap_esp_jax,
                                           in_axes=(0, 0, 0, 0, None, None)) # c1, c2, ch1, ch2, alpha, lam

        expected_np_b = []
        for i in range(batch_size):
            expected_np_b.append(es_np.get_overlap_esp_np(c1_np_b[i], c2_np_b[i], ch1_np_b[i], ch2_np_b[i], alpha, lam))
        expected_np_array = np.array(expected_np_b)

        actual_jax_b = vmapped_get_overlap_jax(c1_jnp_b, c2_jnp_b, ch1_jnp_b, ch2_jnp_b, alpha, lam)
        np.testing.assert_allclose(np.array(actual_jax_b), expected_np_array, rtol=1e-5)

    def test_esp_comparison_jax_vmap_batch(self):
        lam = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        batch_size = 2
        data = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=204)

        points_1_np_b = data[8]
        centers_w_H_2_np_b = data[5]
        partial_charges_2_np_b = data[7]
        point_charges_1_np_b = data[10]
        radii_2_np_b = data[13]

        points_1_jnp_b = data[36]
        centers_w_H_2_jnp_b = data[33]
        partial_charges_2_jnp_b = data[35]
        point_charges_1_jnp_b = data[38]
        radii_2_jnp_b = data[41]

        vmapped_esp_comp_jax = jax.vmap(es_jax._esp_comparison_jax,
                                        in_axes=(0, 0, 0, 0, 0, None, None)) # p1, c_w_H2, pc2, pc1, r2, probe, lam

        expected_np_b = []
        for i in range(batch_size):
            expected_np_b.append(es_np._esp_comparison_np(points_1_np_b[i], centers_w_H_2_np_b[i],
                                                           partial_charges_2_np_b[i], point_charges_1_np_b[i],
                                                           radii_2_np_b[i], probe_radius, lam))
        expected_np_array = np.array(expected_np_b)

        actual_jax_b = vmapped_esp_comp_jax(points_1_jnp_b, centers_w_H_2_jnp_b, partial_charges_2_jnp_b,
                                            point_charges_1_jnp_b, radii_2_jnp_b, probe_radius, lam)
        np.testing.assert_allclose(np.array(actual_jax_b), expected_np_array, rtol=1e-5)

    def test_esp_combo_score_jax_vmap_batch(self):
        alpha = ALPHA_ESP
        lam_comp = LAM_ESP_COMPARISON
        probe_radius = PROBE_RADIUS
        esp_weight = ESP_WEIGHT
        batch_size = 2

        data_b = _generate_electrostatic_data(batch_size=batch_size, fixed_seed=205, num_centers1=5, num_centers2=6)
        (centers_overlap_1_np_b, centers_overlap_2_np_b, _, _,
        centers_w_H_1_np_b, centers_w_H_2_np_b, partial_charges_1_np_b, partial_charges_2_np_b,
        points_1_np_b, points_2_np_b, point_charges_1_np_b, point_charges_2_np_b,
        radii_1_np_b, radii_2_np_b,
        _,_,_,_,_,_,_,_,_,_,_,_,_,_,
        centers_overlap_1_jnp_b, centers_overlap_2_jnp_b, _, _,
        centers_w_H_1_jnp_b, centers_w_H_2_jnp_b, partial_charges_1_jnp_b, partial_charges_2_jnp_b,
        points_1_jnp_b, points_2_jnp_b, point_charges_1_jnp_b, point_charges_2_jnp_b,
        radii_1_jnp_b, radii_2_jnp_b) = data_b

        vmapped_combo_jax = jax.vmap(es_jax.esp_combo_score_jax,
                                     in_axes=(0,0,0,0,0,0,0,0,0,0,0,0, None, None, None, None))

        expected_np_b = []
        for i in range(batch_size):
            expected_np_b.append(
                es_np.esp_combo_score_np(centers_w_H_1_np_b[i], centers_w_H_2_np_b[i],
                                         centers_overlap_1_np_b[i], centers_overlap_2_np_b[i],
                                         points_1_np_b[i], points_2_np_b[i],
                                         partial_charges_1_np_b[i], partial_charges_2_np_b[i],
                                         point_charges_1_np_b[i], point_charges_2_np_b[i],
                                         radii_1_np_b[i], radii_2_np_b[i],
                                         alpha, lam_comp, probe_radius, esp_weight)
            )
        expected_np_array = np.array(expected_np_b)

        actual_jax_b = vmapped_combo_jax(centers_w_H_1_jnp_b, centers_w_H_2_jnp_b,
                                           centers_overlap_1_jnp_b, centers_overlap_2_jnp_b,
                                           points_1_jnp_b, points_2_jnp_b,
                                           partial_charges_1_jnp_b, partial_charges_2_jnp_b,
                                           point_charges_1_jnp_b, point_charges_2_jnp_b,
                                           radii_1_jnp_b, radii_2_jnp_b,
                                           alpha, lam_comp, probe_radius, esp_weight)
        np.testing.assert_allclose(np.array(actual_jax_b), expected_np_array, rtol=1e-5)



"""
Pharmacophore scoring with JAX.

This module may not be as fast as the NumPy version.
"""
from typing import Union, Callable, Literal, Tuple
from functools import partial

import jax.numpy as jnp
from jax import jit, Array

from shepherd_score.score.gaussian_overlap_jax import VAB_2nd_order_jax, VAB_2nd_order_cosine_jax # noqa: F401
from shepherd_score.score.gaussian_overlap_jax import VAB_2nd_order_jax_mask, VAB_2nd_order_cosine_jax_mask
from shepherd_score.score.constants import P_TYPES, P_ALPHAS
P_TYPES_LWRCASE = tuple(map(str.lower, P_TYPES))


# We map string types to integers for JAX scanning
# Mode 0: Scalar (Hydrophobe, Ion) -> effectively sim=1.0
# Mode 1: Vector Abs (Aromatic) -> sim=|dot|
# Mode 2: Vector Clamped (Acceptor, Donor, Halogen) -> sim=max(0, dot)
P_TYPE_CONFIG_MAP = {
    'hydrophobe': {'mode': 0},
    'znbinder':   {'mode': 0},
    'anion':      {'mode': 0},
    'cation':     {'mode': 0},
    'aromatic':   {'mode': 1},
    'acceptor':   {'mode': 2},
    'donor':      {'mode': 2},
    'halogen':    {'mode': 2},
}

SORTED_ALPHAS = []
SORTED_MODES = []

for p_name in P_TYPES:
    key = p_name.lower()
    SORTED_ALPHAS.append(P_ALPHAS[key])
    SORTED_MODES.append(P_TYPE_CONFIG_MAP.get(key, {'mode': 0})['mode'])

TYPE_INDICES = jnp.arange(len(P_TYPES))
LOOKUP_ALPHAS  = jnp.array(SORTED_ALPHAS)
LOOKUP_MODES   = jnp.array(SORTED_MODES)


@partial(jit, static_argnames=['extended_points'])
def precompute_geometry(pos_ref, pos_fit, vec_ref, vec_fit, extended_points=False):
    """
    Computes all necessary geometric matrices once per optimization step.
    """
    # 1. Anchor distances (N, M)
    diff = pos_ref[:, None, :] - pos_fit[None, :, :]
    d2_anchors = jnp.sum(diff**2, axis=-1)

    # 2. Vector dot products (N, M)
    # Safe normalization: eps inside sqrt gives gradient 0 at zero-norm vectors
    # (mode-0 pharmacophores like Hydrophobe/Anion/Cation store [0,0,0] as their
    # direction vector; jnp.linalg.norm has NaN gradient at zero, causing NaN in
    # value_and_grad when fit_vectors_transformed is a zero vector).
    vec_ref_norm = vec_ref / jnp.sqrt(jnp.sum(vec_ref ** 2, axis=-1, keepdims=True) + 1e-12)
    vec_fit_norm = vec_fit / jnp.sqrt(jnp.sum(vec_fit ** 2, axis=-1, keepdims=True) + 1e-12)
    dot_matrix = jnp.einsum('ni,mi->nm', vec_ref_norm, vec_fit_norm)
    # 3. Extended point distances (N, M)
    if extended_points:
        ext_ref = pos_ref + vec_ref # Use original (un-normalized) if they imply length
        ext_fit = pos_fit + vec_fit
        diff_ext = ext_ref[:, None, :] - ext_fit[None, :, :]
        d2_extended = jnp.sum(diff_ext**2, axis=-1)
    else:
        d2_extended = d2_anchors

    return d2_anchors, d2_extended, dot_matrix

@jit
def get_interaction_properties(ptype_idxs):
    """
    Retrieves alpha and mode for given pharmacophore indices.
    """
    return LOOKUP_ALPHAS[ptype_idxs], LOOKUP_MODES[ptype_idxs]

@partial(jit, static_argnames=['extended_points', 'only_extended'])
def get_overlap_pharm_jax_vectorized(
    ptype_1: Array, ptype_2: Array,
    anchors_1: Array, anchors_2: Array,
    vectors_1: Array, vectors_2: Array,
    extended_points: bool = False,
    only_extended: bool = False
) -> Array:
    """
    Fully vectorized pharmacophore overlap.

    Parameters
    ----------
    ptype_1 : Array (N,)
    ptype_2 : Array (M,)
    anchors_1 : Array (N,3)
    anchors_2 : Array (M,3)
    vectors_1 : Array (N,3)
    vectors_2 : Array (M,3)
    extended_points : bool
    only_extended : bool

    Returns
    -------
    score : Array (1,)
        Pharmacophore overlap score.
    """
    # 1. Precompute geometry (N, M)
    d2_anchors, d2_ext, dot_matrix = precompute_geometry(
        anchors_1, anchors_2, vectors_1, vectors_2, extended_points
    )

    # 2. Get pharmacophore properties
    # use ptype_1 properties because if ptype_1 != ptype_2, mask is 0 anyway.
    alphas_1, modes_1 = get_interaction_properties(ptype_1)

    # Broadcast to (N, M)
    # alphas_matrix[i, j] = alpha of ref atom i
    alphas_matrix = alphas_1[:, None]
    modes_matrix = modes_1[:, None]
    matches = (ptype_1[:, None] == ptype_2[None, :]) # match mask (N, M)

    # 3. One-shot computation
    prefactors = (jnp.pi / (2.0 * alphas_matrix)) ** 1.5

    # vector similarity for all modes simultaneously using where
    # Note: modes_matrix is (N, 1), dot_matrix is (N, M)
    is_mode_0 = (modes_matrix == 0)
    is_mode_1 = (modes_matrix == 1)
    is_mode_2 = (modes_matrix == 2)

    sim_raw = (
        jnp.where(is_mode_0, 1.0, 0.0) +
        jnp.where(is_mode_1, jnp.abs(dot_matrix), 0.0) +
        jnp.where(is_mode_2, jnp.maximum(0.0, dot_matrix), 0.0)
    )
    vec_factor = (sim_raw + 2.0) / 3.0
    spatial_anchor = jnp.exp(-0.5 * alphas_matrix * d2_anchors)

    # 4. Extended points logic
    if extended_points:
        spatial_extended = jnp.exp(-0.5 * alphas_matrix * d2_ext)

        # If extended_points=True AND mode=2 (Donor/Acc/Halogen), use extended logic
        # Else use standard logic
        use_ext_logic = is_mode_2 # broadcasts to (N, 1)

        # extended score: (SpatialExt + SpatialAnc) * Prefactor
        term_ext_part = spatial_extended
        term_anc_part = jnp.where(only_extended, 0.0, spatial_anchor)
        score_extended = (term_ext_part + term_anc_part) * prefactors

        # standard score: SpatialAnc * VecFactor * Prefactor
        score_standard = spatial_anchor * vec_factor * prefactors

        # select based on type mode
        final_terms = jnp.where(use_ext_logic, score_extended, score_standard)

    else:
        # standard score only
        final_terms = spatial_anchor * vec_factor * prefactors

    # 5. apply mask and sum
    # multiply by matches (0 or 1) to remove invalid pairs
    return jnp.sum(final_terms * matches)


@partial(jit, static_argnames=['overlap_func', 'allow_antiparallel'])
def _compute_ref_overlap_jax(overlap_func: Callable,
                             anchors_1: Array,
                             alpha: float,
                             vectors_1: Union[Array, None] = None,
                             allow_antiparallel: bool = False
                             ) -> Array:
    """ Single instance only. """
    # Just anchor volume overlap
    if (vectors_1 is None):
        VAA = overlap_func(anchors_1, anchors_1, alpha)
    # Anchor and vector volume overlap for single instance
    else:
        VAA = overlap_func(anchors_1, anchors_1, vectors_1, vectors_1, alpha, allow_antiparallel)
    return VAA


@partial(jit, static_argnames=['overlap_func', 'allow_antiparallel'])
def _compute_fit_overlap_jax(overlap_func: Callable,
                             anchors_2: Array,
                             alpha: float,
                             vectors_2: Union[Array, None] = None,
                             allow_antiparallel: bool = False,
                             ) -> Array:
    """ Single instance only. """
    # Just anchor volume overlap
    if (vectors_2 is None):
        VBB = overlap_func(anchors_2, anchors_2, alpha)
    # Anchor and vector volume overlap for single instance
    else:
        VBB = overlap_func(anchors_2, anchors_2, vectors_2, vectors_2, alpha, allow_antiparallel)
    return VBB




@jit
def tanimoto_func_jax(VAB: Array,
                      VAA: Array,
                      VBB: Array
                      ) -> Array:
    """
    Computes Tanimoto similarity.
    Similarity(Tanimoto) = Overlap{1,2} / (Overlap{1,1} + Overlap{2,2} - Overlap{1,2})
    """
    return VAB / (VAA + VBB - VAB)


@jit
def tversky_func_jax(VAB: Array,
                     VAA: Array,
                     VBB: Array,
                     sigma: float
                     ) -> Array:
    """
    Computes Tversky similarity -> clamped to be max of 1.0.
    sigma: [0,1]

    Similarity(Tversky) = Overlap{1,2} / (sigma*Overlap{1,1} + (1-sigma)*Overlap{2,2})
    """
    return jnp.clip(VAB / (sigma * VAA + (1 - sigma) * VBB), min=None, max=1.0)


def get_vector_volume_overlap_score_jax(ptype_str: str,
                                        ptype_1: Array,
                                        ptype_2: Array,
                                        anchors_1: Array,
                                        anchors_2: Array,
                                        vectors_1: Array,
                                        vectors_2: Array,
                                        allow_antiparallel: bool
                                        ) -> Tuple[Array, Array, Array]:
    """
    Compute volumentric overlap score with cosine similarity of vectors. JAX version.
    """
    ptype_str_lwr = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str_lwr)

    mask_1 = ptype_1 == ptype_idx
    mask_2 = ptype_2 == ptype_idx

    VAB = VAB_2nd_order_cosine_jax_mask(anchors_1, anchors_2, vectors_1, vectors_2, mask_1, mask_2, P_ALPHAS[ptype_str_lwr], allow_antiparallel)
    VAA = VAB_2nd_order_cosine_jax_mask(anchors_1, anchors_1, vectors_1, vectors_1, mask_1, mask_1, P_ALPHAS[ptype_str_lwr], allow_antiparallel)
    VBB = VAB_2nd_order_cosine_jax_mask(anchors_2, anchors_2, vectors_2, vectors_2, mask_2, mask_2, P_ALPHAS[ptype_str_lwr], allow_antiparallel)

    return VAB, VAA, VBB


def get_volume_overlap_score_jax(ptype_str: str,
                                 ptype_1: Array,
                                 ptype_2: Array,
                                 anchors_1: Array,
                                 anchors_2: Array
                                 ) -> Tuple[Array, Array, Array]:
    """
    Computes volume overlap score single instance. JAX version.
    """
    ptype_str_lwr = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str_lwr)

    mask_1 = ptype_1 == ptype_idx
    mask_2 = ptype_2 == ptype_idx

    VAB = VAB_2nd_order_jax_mask(anchors_1, anchors_2, mask_1, mask_2, P_ALPHAS[ptype_str_lwr])
    VAA = VAB_2nd_order_jax_mask(anchors_1, anchors_1, mask_1, mask_1, P_ALPHAS[ptype_str_lwr])
    VBB = VAB_2nd_order_jax_mask(anchors_2, anchors_2, mask_2, mask_2, P_ALPHAS[ptype_str_lwr])

    return VAB, VAA, VBB


def get_volume_overlap_score_extended_points_jax(ptype_str: str,
                                                 ptype_1: Array,
                                                 ptype_2: Array,
                                                 anchors_1: Array,
                                                 anchors_2: Array,
                                                 vectors_1: Array,
                                                 vectors_2: Array,
                                                 only_extended: bool = False
                                                 ) -> Tuple[Array, Array, Array]:
    """
    Score both the anchor and extended point volume overlap instead of a vector similarity. JAX version.
    """
    ptype_str_lwr = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str_lwr)
    alpha = P_ALPHAS[ptype_str_lwr]

    mask_1 = ptype_1 == ptype_idx
    mask_2 = ptype_2 == ptype_idx

    VAB, VAA, VBB = jnp.array(0.0), jnp.array(0.0), jnp.array(0.0)

    # Score extended points
    VAB_extended = VAB_2nd_order_jax_mask(anchors_1 + vectors_1, anchors_2 + vectors_2, mask_1, mask_2, alpha)
    VAA_extended = VAB_2nd_order_jax_mask(anchors_1 + vectors_1, anchors_1 + vectors_1, mask_1, mask_1, alpha)
    VBB_extended = VAB_2nd_order_jax_mask(anchors_2 + vectors_2, anchors_2 + vectors_2, mask_2, mask_2, alpha)

    VAB += VAB_extended
    VAA += VAA_extended
    VBB += VBB_extended

    # Score anchors if not only_extended
    if not only_extended:
        VAB_anchor = VAB_2nd_order_jax_mask(anchors_1, anchors_2, mask_1, mask_2, alpha)
        VAA_anchor = VAB_2nd_order_jax_mask(anchors_1, anchors_1, mask_1, mask_1, alpha)
        VBB_anchor = VAB_2nd_order_jax_mask(anchors_2, anchors_2, mask_2, mask_2, alpha)

        VAB += VAB_anchor
        VAA += VAA_anchor
        VBB += VBB_anchor

    return VAB, VAA, VBB


_SIM_TYPE = Literal['tanimoto', 'tversky', 'tversky_ref', 'tversky_fit']

@partial(jit, static_argnames=['similarity', 'extended_points', 'only_extended'])
def get_overlap_pharm_jax(ptype_1: Array,
                          ptype_2: Array,
                          anchors_1: Array,
                          anchors_2: Array,
                          vectors_1: Array,
                          vectors_2: Array,
                          similarity: _SIM_TYPE = 'tanimoto',
                          extended_points: bool = False,
                          only_extended: bool = False
                          ) -> Array:
    """
    JAX implementation to compute pharmacophore score.
    Single instance only.

    Arguments
    ---------
    ptype_1 : Array (N,)
        Indices specifying the pharmacophore type based on order of P_TYPES
    ptype_2 : Array (M,)
        Indices specifying the pharmacophore type based on order of P_TYPES
    anchors_1 : Array (N,3)
        Coordinates for the anchor points of each pharmacophore of molecule 1
    anchors_2 : Array (M,3)
        Coordinates for the anchor points of each pharmacophore of molecule 2
    vectors_1 : Array (N,3)
        Relative unit vectors of each pharmacophore of molecule 1
    vectors_2 : Array (M,3)
        Relative unit vectors of each pharmacophore of molecule 2
    similarity : str
        Specifies what similarity function to use.
    extended_points : bool
        Whether to score HBA/HBD with gaussian overlaps of extended points.
    only_extended : bool
        When `extended_points` is True, decide whether to only score the extended points (ignore anchor overlaps)

    Returns
    -------
    Array (1,)
    """

    # Similarity scoring
    if similarity.lower() == 'tanimoto':
        similarity_func = tanimoto_func_jax
    elif similarity.lower() == 'tversky':
        similarity_func = partial(tversky_func_jax, sigma=0.95)
    elif similarity.lower() == 'tversky_ref':
        similarity_func = partial(tversky_func_jax, sigma=1.)
    elif similarity.lower() == 'tversky_fit':
        similarity_func = partial(tversky_func_jax, sigma=0.05)
    else:
        # JAX jit doesn't like dynamic error raising based on string values well.
        # Consider validating this outside or ensuring `similarity` is from a known set.
        # For now, assume valid input due to _SIM_TYPE or raise error outside jit scope.
        raise ValueError('Argument `similarity` must be one of (tanimoto, tversky, tversky_ref, tversky_fit).')

    overlap, ref_overlap, fit_overlap = jnp.array(0.0), jnp.array(0.0), jnp.array(0.0)

    # Hydrophobe
    VAB_h, VAA_h, VBB_h = get_volume_overlap_score_jax(ptype_str='hydrophobe',
                                                       ptype_1=ptype_1, ptype_2=ptype_2,
                                                       anchors_1=anchors_1, anchors_2=anchors_2)
    overlap += VAB_h
    ref_overlap += VAA_h
    fit_overlap += VBB_h

    # ZnBinder
    VAB_z, VAA_z, VBB_z = get_volume_overlap_score_jax(ptype_str='znbinder',
                                                       ptype_1=ptype_1, ptype_2=ptype_2,
                                                       anchors_1=anchors_1, anchors_2=anchors_2)
    overlap += VAB_z
    ref_overlap += VAA_z
    fit_overlap += VBB_z

    # Anion
    VAB_an, VAA_an, VBB_an = get_volume_overlap_score_jax(ptype_str='anion',
                                                          ptype_1=ptype_1, ptype_2=ptype_2,
                                                          anchors_1=anchors_1, anchors_2=anchors_2)
    overlap += VAB_an
    ref_overlap += VAA_an
    fit_overlap += VBB_an

    # Cation
    VAB_cat, VAA_cat, VBB_cat = get_volume_overlap_score_jax(ptype_str='cation',
                                                             ptype_1=ptype_1, ptype_2=ptype_2,
                                                             anchors_1=anchors_1, anchors_2=anchors_2)
    overlap += VAB_cat
    ref_overlap += VAA_cat
    fit_overlap += VBB_cat

    # Aromatic
    VAB_ar, VAA_ar, VBB_ar = get_vector_volume_overlap_score_jax(ptype_str='aromatic',
                                                                 ptype_1=ptype_1, ptype_2=ptype_2,
                                                                 anchors_1=anchors_1, anchors_2=anchors_2,
                                                                 vectors_1=vectors_1, vectors_2=vectors_2,
                                                                 allow_antiparallel=True)
    overlap += VAB_ar
    ref_overlap += VAA_ar
    fit_overlap += VBB_ar

    # Acceptor
    if extended_points:
        VAB_acc, VAA_acc, VBB_acc = get_volume_overlap_score_extended_points_jax(
                                        ptype_str='acceptor', ptype_1=ptype_1, ptype_2=ptype_2,
                                        anchors_1=anchors_1, anchors_2=anchors_2,
                                        vectors_1=vectors_1, vectors_2=vectors_2,
                                        only_extended=only_extended)
    else:
        VAB_acc, VAA_acc, VBB_acc = get_vector_volume_overlap_score_jax(
                                        ptype_str='acceptor', ptype_1=ptype_1, ptype_2=ptype_2,
                                        anchors_1=anchors_1, anchors_2=anchors_2,
                                        vectors_1=vectors_1, vectors_2=vectors_2,
                                        allow_antiparallel=False)
    overlap += VAB_acc
    ref_overlap += VAA_acc
    fit_overlap += VBB_acc

    # Donor
    if extended_points:
        VAB_don, VAA_don, VBB_don = get_volume_overlap_score_extended_points_jax(
                                        ptype_str='donor', ptype_1=ptype_1, ptype_2=ptype_2,
                                        anchors_1=anchors_1, anchors_2=anchors_2,
                                        vectors_1=vectors_1, vectors_2=vectors_2,
                                        only_extended=only_extended)
    else:
        VAB_don, VAA_don, VBB_don = get_vector_volume_overlap_score_jax(
                                        ptype_str='donor', ptype_1=ptype_1, ptype_2=ptype_2,
                                        anchors_1=anchors_1, anchors_2=anchors_2,
                                        vectors_1=vectors_1, vectors_2=vectors_2,
                                        allow_antiparallel=False)
    overlap += VAB_don
    ref_overlap += VAA_don
    fit_overlap += VBB_don

    # Halogen
    if extended_points:
        VAB_hal, VAA_hal, VBB_hal = get_volume_overlap_score_extended_points_jax(
                                        ptype_str='halogen', ptype_1=ptype_1, ptype_2=ptype_2,
                                        anchors_1=anchors_1, anchors_2=anchors_2,
                                        vectors_1=vectors_1, vectors_2=vectors_2,
                                        only_extended=only_extended)
    else:
        VAB_hal, VAA_hal, VBB_hal = get_vector_volume_overlap_score_jax(
                                    ptype_str='halogen', ptype_1=ptype_1, ptype_2=ptype_2,
                                    anchors_1=anchors_1, anchors_2=anchors_2,
                                    vectors_1=vectors_1, vectors_2=vectors_2,
                                    allow_antiparallel=False)
    overlap += VAB_hal
    ref_overlap += VAA_hal
    fit_overlap += VBB_hal

    scores = similarity_func(VAB=overlap, VAA=ref_overlap, VBB=fit_overlap)
    return scores


@partial(jit, static_argnames=['similarity', 'extended_points', 'only_extended'])
def get_pharm_combo_score_jax(centers_1: Array,
                              centers_2: Array,
                              ptype_1: Array,
                              ptype_2: Array,
                              anchors_1: Array,
                              anchors_2: Array,
                              vectors_1: Array,
                              vectors_2: Array,
                              alpha: float = 0.81,
                              similarity: str = 'tanimoto',
                              extended_points: bool = False,
                              only_extended: bool = False
                              ) -> Array:
    """ Compute a combined shape and pharmacophore score. JAX version. """
    # Similarity scoring
    if similarity.lower() == 'tanimoto':
        similarity_func = tanimoto_func_jax
    elif similarity.lower() == 'tversky':
        similarity_func = partial(tversky_func_jax, sigma=0.95)
    elif similarity.lower() == 'tversky_ref':
        similarity_func = partial(tversky_func_jax, sigma=1.)
    elif similarity.lower() == 'tversky_fit':
        similarity_func = partial(tversky_func_jax, sigma=0.05)
    else:
        raise ValueError('Argument `similarity` must be one of (tanimoto, tversky, tversky_ref, tversky_fit).')

    # Pharmacophore scoring
    pharm_score = get_overlap_pharm_jax(ptype_1=ptype_1,
                                        ptype_2=ptype_2,
                                        anchors_1=anchors_1,
                                        anchors_2=anchors_2,
                                        vectors_1=vectors_1,
                                        vectors_2=vectors_2,
                                        similarity=similarity,
                                        extended_points=extended_points,
                                        only_extended=only_extended)

    # Shape scoring
    VAB_shape = VAB_2nd_order_jax(centers_1=centers_1,
                                  centers_2=centers_2,
                                  alpha=alpha)
    VAA_shape = VAB_2nd_order_jax(centers_1=centers_1,
                                  centers_2=centers_1,
                                  alpha=alpha)
    VBB_shape = VAB_2nd_order_jax(centers_1=centers_2,
                                  centers_2=centers_2,
                                  alpha=alpha)
    shape_score = similarity_func(VAB=VAB_shape,
                                  VAA=VAA_shape,
                                  VBB=VBB_shape)

    score = (pharm_score + shape_score) / 2.0
    return score



"""
Jax Implementations for principal component alignment (pca).

Credit to Lewis J. Martin as this was adapted from
https://github.com/ljmartin/align/blob/main/0.2%20aligning%20principal%20moments%20of%20inertia.ipynb
"""

import jax.numpy as jnp
from jax import vmap, Array
from shepherd_score.alignment_utils.se3_jax import get_SE3_transform_jax, apply_SE3_transform_jax

def compute_moment_of_inertia_jax(points: Array) -> Array:
    """
    Computes the moment of inertia tensor for a set of points. Jax implementation.
    A = x^2 + y^2 + z^2
    B = X^T X
    """
    points = points - jnp.mean(points, axis=0)
    A = jnp.sum(points**2)
    B = points.T @ points
    eye = jnp.eye(3)
    return (A * eye - B) / points.shape[0]


def compute_principal_moments_of_interia_jax(points: Array) -> Array:
    """
    Calculate principal moment of inertia. Jax implementation.
    """
    momint = compute_moment_of_inertia_jax(points)
    eigvals, eigvecs = jnp.linalg.eigh(momint)
    indices = jnp.argsort(-eigvals) #sorting it returns the 'long' axis in index 0.
    # Return transposed which is more intuitive format
    return eigvecs[:, indices].T


def angle_between_vecs_jax(v1: Array, v2: Array) -> Array:
    """ Compute the angle in radians between two vectors. Jax implementation. """
    v1_u = v1 / jnp.linalg.norm(v1)
    v2_u = v2 / jnp.linalg.norm(v2)
    return jnp.arccos(jnp.clip(jnp.dot(v1_u, v2_u), -1.0, 1.0))

vmap_angle_between_vecs_jax = vmap(angle_between_vecs_jax, (0, 0))

vmap_allclose = vmap(jnp.allclose, 0)
def rotation_axis_jax(v1: Array, v2: Array) -> Array:
    """
    Calculate the vector about which to order to rotate `a` to align with `b` (cross product).
    Jax implementation.
    """
    if len(v1.shape) == 2:
        all_close = vmap_allclose(v1, v2)
        same_vectors_idx = jnp.where(all_close)[0]
        v3 = jnp.zeros(v1.shape)
        if same_vectors_idx.size > 0:
            v3 = v3.at[same_vectors_idx].set(jnp.tile(jnp.array([1., 0., 0.]), (len(same_vectors_idx), 1)))
        diff_vectors_idx = jnp.where(not all_close)[0]
        if diff_vectors_idx.size > 0:
            v3 = v3.at[diff_vectors_idx].set(jnp.cross(v1.at[diff_vectors_idx].get(),
                                                       v2.at[diff_vectors_idx].get(), axis=-1))
            v3 = v3.at[diff_vectors_idx].get() / jnp.linalg.norm(v3.at[diff_vectors_idx].get(), axis=1).reshape((-1,1))
    else:
        if jnp.allclose(v1, v2):
            return jnp.array([1., 0., 0.])
        v3 = jnp.cross(v1, v2)
        v3 = v3 / jnp.linalg.norm(v3)
    return v3

def quaternion_from_axis_angle_jax(axis: Array, angle: Array) -> Array:
    """
    Create a Quaternion from a rotation axis and an angle in radians.
    Jax implementation.

    Parameters
    ----------
    axis : Array (3,)
        Axis to rotate about.
    angle: Array (1,)
        Angle in radians.

    Returns
    -------
    quaternion : Array (4,)
    """
    theta = angle / 2.0
    r = jnp.cos(theta)
    i = axis * jnp.sin(theta)
    return jnp.array([r, i[0], i[1], i[2]])

vmap_quaternion_from_axis_angle_jax = vmap(quaternion_from_axis_angle_jax, (0, 0))


def quaternion_mult_jax(p: Array, q: Array) -> Array:
    """
    Multiplication of quaternions p and q. Jax implementation.

    Reference: https://academicflight.com/articles/kinematics/rotation-formalisms/quaternions/

    General use case: The consecutive rotations of q_1 then q_2 is equivalent
    to q_3 = q_2*q_1. (order matters)

    Parameters
    ----------
    p : Array
        The first quaternion with shape (4,).
    q : Array
        The second quaternion with shape (4,).

    Returns
    -------
    Array
        The product of the two quaternions with shape (4,).
    """
    mat1 = jnp.array([[p[0], -p[1], -p[2], -p[3]],
                      [p[1],  p[0], -p[3],  p[2]],
                      [p[2],  p[3],  p[0], -p[1]],
                      [p[3], -p[2],  p[1],  p[0]]])
    pq = mat1 @ q
    return pq

vmap_quaternion_mult_jax = vmap(quaternion_mult_jax, (0, 0))


def quaternions_for_principal_component_alignment_jax(ref_points: Array,
                                                      fit_points: Array
                                                      ) -> Array:
    """
    Computes the 4 quaternions required for alignment of the fit mol along the
    principal components of the reference mol.
    NumPy implementation.

    The computed quaternions assumes that fit_points will be rotated after being centered at COM.

    Parameters
    ----------
    ref_points : Array (N, 3)
        Set of reference points that fit_points will be aligned to.
    fit_points : Array (M, 3)
        Set of points that will be aligned to ref_points.

    Returns
    -------
    quaternions : Array (4, 4)
        Set of four quaternions corresponding to the alignment of fit_points to ref_points in the
        four possible principal component combinations.
    """
    pmi_ref = compute_principal_moments_of_interia_jax(ref_points)

    quaternions = jnp.zeros((4,4))
    for q_index in range(4):
        if q_index == 1:
            # flip orientation of longest axis
            pmi_ref = pmi_ref.at[0].set(-pmi_ref[0])
        elif q_index == 2:
            # unflip orientation of longest axis
            pmi_ref = pmi_ref.at[0].set(-pmi_ref[0])
            # flip orientation of 2nd longest axis
            pmi_ref = pmi_ref.at[1].set(-pmi_ref[1])
        elif q_index == 3:
            # flip orientation of both axes
            pmi_ref = pmi_ref.at[0].set(-pmi_ref[0])

        quat_order = jnp.zeros((2,4))
        # Initially center to COM
        fit_points_adjust = fit_points - jnp.mean(fit_points, axis=0)
        for ax_idx in range(2):
            pmi_fit = compute_principal_moments_of_interia_jax(fit_points_adjust)
            # Angle between principal axis of fit mol and referencne mol
            angle = angle_between_vecs_jax(pmi_fit[ax_idx], pmi_ref[ax_idx])
            # Axis that we are rotating about
            ax = rotation_axis_jax(pmi_fit[ax_idx], pmi_ref[ax_idx])
            # Quaternion
            quat_order = quat_order.at[ax_idx].set(quaternion_from_axis_angle_jax(ax, angle))
            # get SE(3) transformation matrix
            se3_params = jnp.concatenate((quat_order[ax_idx], jnp.zeros(3)))
            # get transformed matrix
            fit_points_adjust = apply_SE3_transform_jax(fit_points_adjust, get_SE3_transform_jax(se3_params))
        quaternions = quaternions.at[q_index].set(quaternion_mult_jax(quat_order[1], quat_order[0]))
    return quaternions
